import os
import typing as t
from typing import TYPE_CHECKING
from pathlib import Path

import yaml
from simple_di import inject
from simple_di import Provide

import bentoml
from bentoml import Tag
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..utils.pkg import get_pkg_version
from .common.model_runner import BaseModelRunner
from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    from mlflow.pyfunc import PyFuncModel

    from ..models import ModelStore


try:
    import mlflow.pyfunc
    from mlflow.tracking.artifact_utils import _download_artifact_from_uri
except ImportError:
    raise MissingDependencyException(
        """\
        `mlflow` is required to use with `bentoml.mlflow`.
        Instruction: `pip install -U mlflow`
        """
    )

MODULE_NAME = "bentoml.mlflow"


def _strike(text: str) -> str:
    return "".join(["\u0336{}".format(c) for c in text])


def _validate_file_exists(fname: str, parent: str) -> t.Tuple[bool, str]:
    for path in Path(parent).iterdir():
        if path.is_dir():
            return _validate_file_exists(fname, str(path))
        if path.name == fname:
            return True, str(path)
    return False, ""


@inject
def load(
    tag: t.Union[str, Tag],
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "PyFuncModel":
    """
    Load a model from BentoML local modelstore with given name.

    Args:
        tag (:code:`Union[str, Tag]`):
            Tag of a saved model in BentoML local modelstore.
        model_store (:mod:`~bentoml._internal.models.store.ModelStore`, default to :mod:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj:`mlflow.pyfunc.PyFuncModel`: an instance of `mlflow.pyfunc.PyFuncModel` from BentoML modelstore.

    Examples:

    .. code-block:: python

        import bentoml

        model = bentoml.mlflow.load("mlflow_sklearn_model")

    """  # noqa
    model = model_store.get(tag)
    if model.info.module not in (MODULE_NAME, __name__):
        raise BentoMLException(
            f"Model {tag} was saved with module {model.info.module}, failed loading with {MODULE_NAME}."
        )
    mlflow_folder = model.path_of(model.info.options["mlflow_folder"])
    return mlflow.pyfunc.load_model(mlflow_folder, suppress_warnings=False)  # type: ignore


SAVE_WARNING = f"""\
BentoML won't provide a :func:`save` API for MLflow. If one uses :code:`bentoml.mlflow.save`, it will
raises :obj:`BentoMLException`:

    - If you currently working with :code:`mlflow.<flavor>.save_model`, we kindly suggest you
      to replace :code:`mlflow.<flavor>.save_model` with BentoML's `save` API as we also supports
      the majority of ML frameworks that MLflow supports. An example below shows how you can
      migrate your code for PyTorch from MLflow to BentoML:

    .. code-block:: diff

        - {_strike('import mlflow.pytorch')}
        + import bentoml

        # PyTorch model logics
        ...
        model = EmbeddingBag()
        - {_strike('mlflow.pytorch.save_model(model, "embed_bag")')}
        + bentoml.pytorch.save("embed_bag", model)

    - If you want to import MLflow models from local directory or a given file path, you can
      utilize :code:`bentoml.mlflow.import_from_uri`:

    .. code-block:: diff

        import mlflow.pytorch
        + import bentoml

        path = "./my_pytorch_model"
        mlflow.pytorch.save_model(model, path)
        + tag = bentoml.mlflow.import_from_uri("mlflow_pytorch_model", path)

    - If your current workflow with MLflow involve :func:`log_model` as well as importing models from MLflow Registry,
      you can import those directly to BentoML modelstore using :code:`bentoml.mlflow.import_from_uri`. We also accept
      MLflow runs syntax, as well as models registry uri.
      An example showing how to integrate your current :func:`log_model` with :code:`mlflow.sklearn` to BentoML:

    .. code-block:: diff

        import mlflow.sklearn
        + import mlflow
        + import bentoml

        # Log sklearn model `sk_learn_rfr` and register as version 1
        ...
        reg_name = "sk-learn-random-forest-reg-model"
        artifact_path = "sklearn_model"
        mlflow.sklearn.log_model(
            sk_model=sk_learn_rfr,
            artifact_path=artifact_path,
            registered_model_name=reg_name
        )

        # refers to https://www.mlflow.org/docs/latest/tracking.html#logging-functions
        + current_run = mlflow.active_run().info.run_id
        + uri = "runs:/%s/%s" % (current_run, artifact_path)
        + tag = bentoml.mlflow.import_from_uri("runs_mlflow_sklearn", uri)

An example showing how to import from MLflow models registry to BentoML modelstore. With this usecase, we
recommend you to load the model into memory first with :code:`mlflow.<flavor>.load_model` then save the model using
BentoML :func:`save` API:

.. code-block:: diff

    import mlflow.sklearn
    + import bentoml

    reg_model_name = "sk-learn-random-forest-reg-model"
    model_uri = "models:/%s/1" % reg_model_name
    + loaded = mlflow.sklearn.load_model(model_uri, *args, **kwargs)
    + tag = bentoml.sklearn.save("my_model", loaded, *args, **kwargs)

    # you can also use `bentoml.mlflow.import_from_uri` to import the model directly after
    #  defining `model_uri`
    + import bentoml
    + tag = bentoml.mlflow.import_from_uri("my_model", model_uri)

"""  # noqa


def save(*args: str, **kwargs: str) -> None:  # noqa # pylint: disable
    raise BentoMLException(SAVE_WARNING)


save.__doc__ = SAVE_WARNING


def import_from_uri(
    name: str,
    uri: str,
    *,
    labels: t.Optional[t.Dict[str, str]] = None,
    custom_objects: t.Optional[t.Dict[str, t.Any]] = None,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
) -> Tag:
    """
    Imports a MLFlow model format to BentoML modelstore via given URI.

    Args:
        name (:code:`str`):
            Name for your MLFlow model to be saved under BentoML modelstore.
        uri (:code:`str`): URI accepts all MLflow defined APIs in terms of referencing artifacts.
            All available accepted URI (extracted from `MLFlow Concept <https://mlflow.org/docs/latest/concepts.html>`_):
                - :code:`/Users/me/path/to/local/model`
                - :code:`relative/path/to/local/model`
                - :code:`s3://my_bucket/path/to/model`
                - :code:`hdfs://<host>:<port>/<path>`
                - :code:`runs:/<mlflow_run_id>/run-relative/path/to/model`
                - :code:`models:/<model_name>/<model_version>`
                - :code:`models:/<model_name>/<stage>`
        labels (:code:`Dict[str, str]`, `optional`, default to :code:`None`):
            user-defined labels for managing models, e.g. team=nlp, stage=dev
        custom_objects (:code:`Dict[str, Any]]`, `optional`, default to :code:`None`):
            user-defined additional python objects to be saved alongside the model,
            e.g. a tokenizer instance, preprocessor function, model configuration json
        metadata (:code:`Dict[str, Any]`, `optional`, default to :code:`None`):
            Custom metadata for given model.
        model_store (:mod:`~bentoml._internal.models.store.ModelStore`, default to :mod:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`~bentoml.Tag` object that can be used to retrieve the model with :func:`bentoml.tensorflow.load`:

    Example:

    .. code-block:: python

        from sklearn import svm, datasets

        import mlflow
        import bentoml

        # Load training data
        iris = datasets.load_iris()
        X, y = iris.data, iris.target

        # Model Training
        clf = svm.SVC()
        clf.fit(X, y)

        # Wrap up as a custom pyfunc model
        class ModelPyfunc(mlflow.pyfunc.PythonModel):

            def load_context(self, context):
                self.model = clf

            def predict(self, context, model_input):
                return self.model.predict(model_input)

        # Log model
        with mlflow.start_run() as run:
            model = ModelPyfunc()
            mlflow.pyfunc.log_model("model", python_model=model)
            print("run_id: {}".format(run.info.run_id))

        model_uri = f"runs:/{run.info.run_id}/model"

        # import given model from local uri to BentoML modelstore:
        tag = bentoml.mlflow.import_from_uri("model", model_uri)

        # from MLFlow Models API
        model_tag = bentoml.mlflow.import_from_uri("mymodel", "models:/mymodel/1")
    """  # noqa
    context: t.Dict[str, t.Any] = {
        "framework_name": "mlflow",
        "pip_dependencies": [f"mlflow=={get_pkg_version('mlflow')}"],
    }

    with bentoml.models.create(
        name,
        module=MODULE_NAME,
        labels=labels,
        custom_objects=custom_objects,
        options=None,
        context=context,
        metadata=metadata,
    ) as _model:

        _model.info.options = {"uri": uri}
        mlflow_obj_path = _download_artifact_from_uri(uri, output_path=_model.path)
        _model.info.options["mlflow_folder"] = os.path.relpath(
            mlflow_obj_path, _model.path
        )
        exists, _ = _validate_file_exists("MLproject", mlflow_obj_path)
        if exists:
            raise BentoMLException("BentoML doesn't accept MLflow Projects.")
        exists, fpath = _validate_file_exists("MLmodel", mlflow_obj_path)
        if not exists:
            raise BentoMLException("Downloaded path is not a valid MLflow Model.")
        with Path(fpath).open("r") as mf:
            conf = yaml.safe_load(mf.read())
            flavor = conf["flavors"]["python_function"]
            _model.info.options["flavor"] = flavor["loader_module"]

        return _model.tag


class _PyFuncRunner(BaseModelRunner):
    @property
    def num_replica(self) -> int:
        return 1

    def _setup(self) -> None:
        self._model = load(self._tag, self.model_store)

    def _run_batch(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        return self._model.predict(*args, **kwargs)  # type: ignore


@inject
def load_runner(
    tag: t.Union[str, Tag],
    name: t.Optional[str] = None,
) -> "_PyFuncRunner":
    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. `bentoml.mlflow.load_runner` implements a Runner class that
    has to options to wrap around a PyFuncModel, or infer from given MLflow flavor to
    load BentoML internal Runner implementation, which optimize it for the BentoML runtime.

    Args:
        tag (:code:`Union[str, Tag]`):
            Tag of a saved model in BentoML local modelstore.

    Returns:
        :obj:`bentoml._internal.runner.Runner`: Runner instances loaded from `bentoml.mlflow`.

    .. note::


       Currently this is an instance of :obj:`~bentoml._internal.frameworks.mlflow._PyFuncRunner` which is the base
       runner. We recommend users to use the correspond frameworks' Runner implementation provided by BentoML for the best performance.

    On our roadmap, the intention for this API is to load the coresponding framework runner automatically. For example:

    .. code-block:: python

        import bentoml

        # this tag `mlflow_pytorch` is imported from mlflow
        # when user use `bentoml.mlflow.load_runner`, it should returns
        # bentoml._internal.frameworks.pytorch.PyTorchRunner` instead
        # and kwargs can be passed directly to `PyTorchRunner`
        runner = bentoml.mlflow.load_runner("mlflow_pytorch", **kwargs)

    Examples:

    .. code-block:: python

        import bentoml

        runner = bentoml.mlflow.load_runner(tag)
        runner.run_batch([[1,2,3,4,5]])

    """
    return _PyFuncRunner(
        tag,
        name=name,
    )



import re
import uuid
import base64
import typing as t
import logging

import fs
import attr

from .utils import bentoml_cattr
from ..exceptions import InvalidArgument
from ..exceptions import BentoMLException

logger = logging.getLogger(__name__)

tag_fmt = "[a-z0-9]([-._a-z0-9]*[a-z0-9])?"
tag_max_length = 63
tag_max_length_error_msg = (
    "a tag's name or version must be at most {tag_max_length} characters in length"
)
tag_invalid_error_msg = "a tag's name or version must consist of alphanumeric characters, '_', '-', or '.', and must start and end with an alphanumeric character"
tag_regex = re.compile(f"^{tag_fmt}$")


def validate_tag_str(value: str):
    """
    Validate that a tag value (either name or version) is a simple string that:
        * Must be at most 63 characters long,
        * Begin and end with an alphanumeric character (`[a-z0-9A-Z]`), and
        * May contain dashes (`-`), underscores (`_`) dots (`.`), or alphanumerics
          between.
    """
    errors = []
    if len(value) > tag_max_length:
        errors.append(tag_max_length_error_msg)
    if tag_regex.match(value) is None:
        errors.append(tag_invalid_error_msg)

    if errors:
        raise InvalidArgument(f"{value} is not a valid tag: " + ", and ".join(errors))


@attr.define(slots=True)
class Tag:
    name: str
    version: t.Optional[str]

    def __init__(self, name: str, version: t.Optional[str] = None):
        lname = name.lower()
        if name != lname:
            logger.warning(f"converting {name} to lowercase: {lname}")

        validate_tag_str(lname)

        self.name = lname

        if version is not None:
            lversion = version.lower()
            if version != lversion:
                logger.warning(f"converting {version} to lowercase: {lversion}")
            validate_tag_str(lversion)
            self.version = lversion
        else:
            self.version = None

    def __str__(self):
        if self.version is None:
            return self.name
        else:
            return f"{self.name}:{self.version}"

    def __repr__(self):
        return f"{self.__class__.__name__}(name={repr(self.name)}, version={repr(self.version)})"

    def __eq__(self, other: "Tag") -> bool:
        return self.name == other.name and self.version == other.version

    def __lt__(self, other: "Tag") -> bool:
        if self.name == other.name:
            if other.version is None:
                return False
            if self.version is None:
                return True
            return self.version < other.version
        return self.name < other.name

    def __hash__(self) -> int:
        return hash((self.name, self.version))

    @classmethod
    def from_taglike(cls, taglike: t.Union["Tag", str]) -> "Tag":
        if isinstance(taglike, Tag):
            return taglike
        return cls.from_str(taglike)

    @classmethod
    def from_str(cls, tag_str: str) -> "Tag":
        if ":" not in tag_str:
            return cls(tag_str, None)
        try:
            name, _, version = tag_str.partition(":")
            if not version:
                # in case users mistakenly define "bento:"
                raise BentoMLException(
                    f"{tag_str} contains trailing ':'. Maybe you meant to use `{tag_str}:latest`?"
                )
            return cls(name, version)
        except ValueError:
            raise BentoMLException(f"Invalid {cls.__name__} {tag_str}")

    def make_new_version(self) -> "Tag":
        if self.version is not None:
            raise ValueError(
                "tried to run 'make_new_version' on a Tag that already has a version"
            )
        ver_bytes = bytearray(uuid.uuid1().bytes)
        # cut out the time_hi and node bits of the uuid
        ver_bytes = ver_bytes[:6] + ver_bytes[8:12]
        encoded_ver = base64.b32encode(ver_bytes)

        return Tag(self.name, encoded_ver.decode("ascii").lower())

    def path(self) -> str:
        if self.version is None:
            return self.name
        return fs.path.combine(self.name, self.version)

    def latest_path(self) -> str:
        return fs.path.combine(self.name, "latest")


# Remove after attrs support ForwardRef natively
attr.resolve_types(Tag, globals(), locals())

bentoml_cattr.register_structure_hook(Tag, lambda d, _: Tag.from_taglike(d))  # type: ignore[misc]
bentoml_cattr.register_unstructure_hook(Tag, lambda tag: str(tag))



"""
User facing python APIs for managing local bentos and build new bentos
"""

from __future__ import annotations

import os
import typing as t
import logging
import subprocess
from typing import TYPE_CHECKING

from simple_di import inject
from simple_di import Provide

from bentoml.exceptions import InvalidArgument

from ._internal.tag import Tag
from ._internal.bento import Bento
from ._internal.utils import resolve_user_filepath
from ._internal.bento.build_config import BentoBuildConfig
from ._internal.configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    from ._internal.bento import BentoStore
    from ._internal.types import PathType
    from ._internal.models import ModelStore

logger = logging.getLogger(__name__)

BENTOML_FIGLET = """
██████╗░███████╗███╗░░██╗████████╗░█████╗░███╗░░░███╗██╗░░░░░
██╔══██╗██╔════╝████╗░██║╚══██╔══╝██╔══██╗████╗░████║██║░░░░░
██████╦╝█████╗░░██╔██╗██║░░░██║░░░██║░░██║██╔████╔██║██║░░░░░
██╔══██╗██╔══╝░░██║╚████║░░░██║░░░██║░░██║██║╚██╔╝██║██║░░░░░
██████╦╝███████╗██║░╚███║░░░██║░░░╚█████╔╝██║░╚═╝░██║███████╗
╚═════╝░╚══════╝╚═╝░░╚══╝░░░╚═╝░░░░╚════╝░╚═╝░░░░░╚═╝╚══════╝
"""


@inject
def list(  # pylint: disable=redefined-builtin
    tag: t.Optional[t.Union[Tag, str]] = None,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
) -> "t.List[Bento]":
    return _bento_store.list(tag)


@inject
def get(
    tag: t.Union[Tag, str],
    *,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
) -> Bento:
    return _bento_store.get(tag)


@inject
def delete(
    tag: t.Union[Tag, str],
    *,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
):
    _bento_store.delete(tag)


@inject
def import_bento(
    path: str,
    input_format: t.Optional[str] = None,
    *,
    protocol: t.Optional[str] = None,
    user: t.Optional[str] = None,
    passwd: t.Optional[str] = None,
    params: t.Optional[t.Dict[str, str]] = None,
    subpath: t.Optional[str] = None,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
) -> Bento:
    """
    Import a bento.

    Examples:

    .. code-block:: python

        # imports 'my_bento' from '/path/to/folder/my_bento.bento'
        bentoml.import_bento('/path/to/folder/my_bento.bento')

        # imports 'my_bento' from '/path/to/folder/my_bento.tar.gz'
        # currently supported formats are tar.gz ('gz'), tar.xz ('xz'), tar.bz2 ('bz2'), and zip
        bentoml.import_bento('/path/to/folder/my_bento.tar.gz')
        # treats 'my_bento.ext' as a gzipped tarfile
        bentoml.import_bento('/path/to/folder/my_bento.ext', 'gz')

        # imports 'my_bento', which is stored as an uncompressed folder, from '/path/to/folder/my_bento/'
        bentoml.import_bento('/path/to/folder/my_bento', 'folder')

        # imports 'my_bento' from the S3 bucket 'my_bucket', path 'folder/my_bento.bento'
        # requires `fs-s3fs <https://pypi.org/project/fs-s3fs/>`_ ('pip install fs-s3fs')
        bentoml.import_bento('s3://my_bucket/folder/my_bento.bento')
        bentoml.import_bento('my_bucket/folder/my_bento.bento', protocol='s3')
        bentoml.import_bento('my_bucket', protocol='s3', subpath='folder/my_bento.bento')
        bentoml.import_bento('my_bucket', protocol='s3', subpath='folder/my_bento.bento',
                             user='<AWS access key>', passwd='<AWS secret key>',
                             params={'acl': 'public-read', 'cache-control': 'max-age=2592000,public'})

    For a more comprehensive description of what each of the keyword arguments (:code:`protocol`,
    :code:`user`, :code:`passwd`, :code:`params`, and :code:`subpath`) mean, see the
    `FS URL documentation <https://docs.pyfilesystem.org/en/latest/openers.html>`_.

    Args:
        tag: the tag of the bento to export
        path: can be one of two things:
            * a folder on the local filesystem
            * an `FS URL <https://docs.pyfilesystem.org/en/latest/openers.html>`_, for example
                :code:`'s3://my_bucket/folder/my_bento.bento'`
        protocol: (expert) The FS protocol to use when exporting. Some example protocols are :code:`'ftp'`,
            :code:`'s3'`, and :code:`'userdata'`
        user: (expert) the username used for authentication if required, e.g. for FTP
        passwd: (expert) the username used for authentication if required, e.g. for FTP
        params: (expert) a map of parameters to be passed to the FS used for export, e.g. :code:`{'proxy': 'myproxy.net'}`
            for setting a proxy for FTP
        subpath: (expert) the path inside the FS that the bento should be exported to
        _bento_store: the bento store to save the bento to

    Returns:
        Bento: the imported bento
    """
    return Bento.import_from(
        path,
        input_format,
        protocol=protocol,
        user=user,
        passwd=passwd,
        params=params,
        subpath=subpath,
    ).save(_bento_store)


@inject
def export_bento(
    tag: t.Union[Tag, str],
    path: str,
    output_format: t.Optional[str] = None,
    *,
    protocol: t.Optional[str] = None,
    user: t.Optional[str] = None,
    passwd: t.Optional[str] = None,
    params: t.Optional[t.Dict[str, str]] = None,
    subpath: t.Optional[str] = None,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
) -> str:
    """
    Export a bento.

    Examples:

    .. code-block:: python

        # exports 'my_bento' to '/path/to/folder/my_bento-version.bento' in BentoML's default format
        bentoml.export_bento('my_bento:latest', '/path/to/folder')
        # note that folders can only be passed if exporting to the local filesystem; otherwise the
        # full path, including the desired filename, must be passed

        # exports 'my_bento' to '/path/to/folder/my_bento.bento' in BentoML's default format
        bentoml.export_bento('my_bento:latest', '/path/to/folder/my_bento')
        bentoml.export_bento('my_bento:latest', '/path/to/folder/my_bento.bento')

        # exports 'my_bento' to '/path/to/folder/my_bento.tar.gz' in gzip format
        # currently supported formats are tar.gz ('gz'), tar.xz ('xz'), tar.bz2 ('bz2'), and zip
        bentoml.export_bento('my_bento:latest', '/path/to/folder/my_bento.tar.gz')
        # outputs a gzipped tarfile as 'my_bento.ext'
        bentoml.export_bento('my_bento:latest', '/path/to/folder/my_bento.ext', 'gz')

        # exports 'my_bento' to '/path/to/folder/my_bento/' as a folder
        bentoml.export_bento('my_bento:latest', '/path/to/folder/my_bento', 'folder')

        # exports 'my_bento' to the S3 bucket 'my_bucket' as 'folder/my_bento-version.bento'
        bentoml.export_bento('my_bento:latest', 's3://my_bucket/folder')
        bentoml.export_bento('my_bento:latest', 'my_bucket/folder', protocol='s3')
        bentoml.export_bento('my_bento:latest', 'my_bucket', protocol='s3', subpath='folder')
        bentoml.export_bento('my_bento:latest', 'my_bucket', protocol='s3', subpath='folder',
                             user='<AWS access key>', passwd='<AWS secret key>',
                             params={'acl': 'public-read', 'cache-control': 'max-age=2592000,public'})

    For a more comprehensive description of what each of the keyword arguments (:code:`protocol`,
    :code:`user`, :code:`passwd`, :code:`params`, and :code:`subpath`) mean, see the
    `FS URL documentation <https://docs.pyfilesystem.org/en/latest/openers.html>`_.

    Args:
        tag: the tag of the Bento to export
        path: can be one of two things:
            * a folder on the local filesystem
            * an `FS URL <https://docs.pyfilesystem.org/en/latest/openers.html>`_
                * for example, :code:`'s3://my_bucket/folder/my_bento.bento'`
        protocol: (expert) The FS protocol to use when exporting. Some example protocols are :code:`'ftp'`,
            :code:`'s3'`, and :code:`'userdata'`
        user: (expert) the username used for authentication if required, e.g. for FTP
        passwd: (expert) the username used for authentication if required, e.g. for FTP
        params: (expert) a map of parameters to be passed to the FS used for export, e.g. :code:`{'proxy': 'myproxy.net'}`
            for setting a proxy for FTP
        subpath: (expert) the path inside the FS that the bento should be exported to
        _bento_store: save Bento created to this BentoStore

    Returns:
        str: A representation of the path that the Bento was exported to. If it was exported to the local filesystem,
            this will be the OS path to the exported Bento. Otherwise, it will be an FS URL.
    """
    bento = get(tag, _bento_store=_bento_store)
    return bento.export(
        path,
        output_format,
        protocol=protocol,
        user=user,
        passwd=passwd,
        params=params,
        subpath=subpath,
    )


@inject
def push(
    tag: t.Union[Tag, str],
    *,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
):
    raise NotImplementedError


@inject
def pull(
    tag: t.Union[Tag, str],
    *,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
):
    raise NotImplementedError


@inject
def build(
    service: str,
    *,
    labels: t.Optional[t.Dict[str, str]] = None,
    description: t.Optional[str] = None,
    include: t.Optional[t.List[str]] = None,
    exclude: t.Optional[t.List[str]] = None,
    docker: t.Optional[t.Dict[str, t.Any]] = None,
    python: t.Optional[t.Dict[str, t.Any]] = None,
    conda: t.Optional[t.Dict[str, t.Any]] = None,
    version: t.Optional[str] = None,
    build_ctx: t.Optional[str] = None,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
    _model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "Bento":
    """
    User-facing API for building a Bento. The available build options are identical to the keys of a
    valid 'bentofile.yaml' file.

    This API will not respect any 'bentofile.yaml' files. Build options should instead be provided
    via function call parameters.

    Args:
        service: import str for finding the bentoml.Service instance build target
        labels: optional immutable labels for carrying contextual info
        description: optional description string in markdown format
        include: list of file paths and patterns specifying files to include in Bento,
            default is all files under build_ctx, beside the ones excluded from the
            exclude parameter or a :code:`.bentoignore` file for a given directory
        exclude: list of file paths and patterns to exclude from the final Bento archive
        docker: dictionary for configuring Bento's containerization process, see details
            in :class:`bentoml._internal.bento.build_config.DockerOptions`
        python: dictionary for configuring Bento's python dependencies, see details in
            :class:`bentoml._internal.bento.build_config.PythonOptions`
        conda: dictionary for configuring Bento's conda dependencies, see details in
            :class:`bentoml._internal.bento.build_config.CondaOptions`
        version: Override the default auto generated version str
        build_ctx: Build context directory, when used as
        _bento_store: save Bento created to this BentoStore
        _model_store: pull Models required from this ModelStore

    Returns:
        Bento: a Bento instance representing the materialized Bento saved in BentoStore

    Example:

        .. code-block::
            import bentoml

            bentoml.build(
                service="fraud_detector.py:svc",
                version="any_version_label",  # override default version generator
                description=open("README.md").read(),
                include=['*'],
                exclude=[], # files to exclude can also be specified with a .bentoignore file
                labels={
                    "foo": "bar",
                    "team": "abc"
                },
                python=dict(
                    packages=["tensorflow", "numpy"],
                    # requirements_txt="./requirements.txt",
                    index_url="http://<api token>:@mycompany.com/pypi/simple",
                    trusted_host=["mycompany.com"],
                    find_links=['thirdparty..'],
                    extra_index_url=["..."],
                    pip_args="ANY ADDITIONAL PIP INSTALL ARGS",
                    wheels=["./wheels/*"],
                    lock_packages=True,
                ),
                docker=dict(
                    distro="amazonlinux2",
                    setup_script="setup_docker_container.sh",
                    python_version="3.8",
                ),
            )

    """  # noqa: LN001
    build_config = BentoBuildConfig(
        service=service,
        description=description,
        labels=labels,
        include=include,
        exclude=exclude,
        docker=docker,  # type: ignore
        python=python,  # type: ignore
        conda=conda,  # type: ignore
    )

    bento = Bento.create(
        build_config=build_config,
        version=version,
        build_ctx=build_ctx,
    ).save(_bento_store)
    logger.info(BENTOML_FIGLET)
    logger.info('Successfully built %s at "%s"', bento, bento.path)
    return bento


@inject
def build_bentofile(
    bentofile: str = "bentofile.yaml",
    *,
    version: t.Optional[str] = None,
    build_ctx: t.Optional[str] = None,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
    _model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "Bento":
    """
    Build a Bento base on options specified in a bentofile.yaml file.

    By default, this function will look for a `bentofile.yaml` file in current working
    directory.

    Args:
        bentofile: The file path to build config yaml file
        version: Override the default auto generated version str
        build_ctx: Build context directory, when used as
        _bento_store: save Bento created to this BentoStore
        _model_store: pull Models required from this ModelStore
    """
    try:
        bentofile = resolve_user_filepath(bentofile, build_ctx)
    except FileNotFoundError:
        raise InvalidArgument(f'bentofile "{bentofile}" not found')

    with open(bentofile, "r", encoding="utf-8") as f:
        build_config = BentoBuildConfig.from_yaml(f)

    bento = Bento.create(
        build_config=build_config,
        version=version,
        build_ctx=build_ctx,
    ).save(_bento_store)
    logger.info(BENTOML_FIGLET)
    logger.info('Successfully built %s at "%s"', bento, bento.path)
    return bento


@inject
def containerize(
    tag: Tag | str,
    docker_image_tag: str | None = None,
    *,
    add_host: dict[str, str] | None = None,
    allow: t.List[str] | None = None,
    build_args: dict[str, str] | None = None,
    build_context: dict[str, str] | None = None,
    builder: str | None = None,
    cache_from: str | t.List[str] | dict[str, str] | None = None,
    cache_to: str | t.List[str] | dict[str, str] | None = None,
    cgroup_parent: str | None = None,
    iidfile: PathType | None = None,
    labels: dict[str, str] | None = None,
    load: bool = True,
    metadata_file: PathType | None = None,
    network: str | None = None,
    no_cache: bool = False,
    no_cache_filter: t.List[str] | None = None,
    output: str | dict[str, str] | None = None,
    platform: str | t.List[str] | None = None,
    progress: t.Literal["auto", "tty", "plain"] = "auto",
    pull: bool = False,
    push: bool = False,
    quiet: bool = False,
    secrets: str | t.List[str] | None = None,
    shm_size: str | int | None = None,
    rm: bool = False,
    ssh: str | None = None,
    target: str | None = None,
    ulimit: str | None = None,
    _bento_store: "BentoStore" = Provide[BentoMLContainer.bento_store],
) -> bool:

    from bentoml._internal.utils import buildx

    env = {"DOCKER_BUILDKIT": "1", "DOCKER_SCAN_SUGGEST": "false"}

    # run health check whether buildx is install locally
    buildx.health()

    bento = _bento_store.get(tag)
    if docker_image_tag is None:
        docker_image_tag = str(bento.tag)

    dockerfile_path = os.path.join("env", "docker", "Dockerfile")

    logger.info(f"Building docker image for {bento}...")
    try:
        buildx.build(
            subprocess_env=env,
            cwd=bento.path,
            file=dockerfile_path,
            tags=docker_image_tag,
            add_host=add_host,
            allow=allow,
            build_args=build_args,
            build_context=build_context,
            builder=builder,
            cache_from=cache_from,
            cache_to=cache_to,
            cgroup_parent=cgroup_parent,
            iidfile=iidfile,
            labels=labels,
            load=load,
            metadata_file=metadata_file,
            network=network,
            no_cache=no_cache,
            no_cache_filter=no_cache_filter,
            output=output,
            platform=platform,
            progress=progress,
            pull=pull,
            push=push,
            quiet=quiet,
            secrets=secrets,
            shm_size=shm_size,
            rm=rm,
            ssh=ssh,
            target=target,
            ulimit=ulimit,
        )
    except subprocess.CalledProcessError as e:
        logger.error(f"Failed building docker image: {e}")
        if platform != "linux/amd64":
            logger.debug(
                f"""If you run into the following error: "failed to solve: pull access denied, repository does not exist or may require authorization: server message: insufficient_scope: authorization failed". This means Docker doesn't have context of your build platform {platform}. By default BentoML will set target build platform to the current machine platform via `uname -m`. Try again by specifying to build x86_64 (amd64) platform: bentoml containerize {str(bento.tag)} --platform linux/amd64"""
            )
        return False
    else:
        logger.info(f'Successfully built docker image "{docker_image_tag}"')
        return True


__all__ = [
    "list",
    "get",
    "delete",
    "import_bento",
    "export_bento",
    "push",
    "pull",
    "build",
    "build_bentofile",
    "containerize",
]



from __future__ import annotations

import typing as t
from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING

from starlette.requests import Request
from starlette.responses import Response

if TYPE_CHECKING:
    from types import UnionType

    from ..types import LazyType
    from ..context import InferenceApiContext as Context


IOPyObj = t.TypeVar("IOPyObj")


_T = t.TypeVar("_T")


class IODescriptor(ABC, t.Generic[IOPyObj]):
    """
    IODescriptor describes the input/output data format of an InferenceAPI defined
    in a :code:`bentoml.Service`. This is an abstract base class for extending new HTTP
    endpoint IO descriptor types in BentoServer.
    """

    HTTP_METHODS = ["POST"]
    _init_str: str = ""

    def __new__(cls: t.Type[_T], *args: t.Any, **kwargs: t.Any) -> _T:
        self = super().__new__(cls)
        arg_strs = tuple(repr(i) for i in args) + tuple(
            f"{k}={repr(v)}" for k, v in kwargs.items()
        )
        setattr(self, "_init_str", f"{cls.__name__}({', '.join(arg_strs)})")

        return self

    def __repr__(self) -> str:
        return self._init_str

    @abstractmethod
    def input_type(
        self,
    ) -> t.Union[
        "UnionType",
        t.Type[t.Any],
        "LazyType[t.Any]",
        t.Dict[str, t.Union[t.Type[t.Any], "UnionType", "LazyType[t.Any]"]],
    ]:
        ...

    @abstractmethod
    def openapi_schema_type(self) -> t.Dict[str, str]:
        ...

    @abstractmethod
    def openapi_request_schema(self) -> t.Dict[str, t.Any]:
        ...

    @abstractmethod
    def openapi_responses_schema(self) -> t.Dict[str, t.Any]:
        ...

    @abstractmethod
    async def from_http_request(self, request: Request) -> IOPyObj:
        ...

    @abstractmethod
    async def to_http_response(
        self, obj: IOPyObj, ctx: Context | None = None
    ) -> Response:
        ...

    # TODO: gRPC support
    # @abstractmethod
    # def generate_protobuf(self): ...

    # @abstractmethod
    # async def from_grpc_request(self, request: GRPCRequest) -> IOType: ...

    # @abstractmethod
    # async def to_grpc_response(self, obj: IOType) -> GRPCResponse: ...



from __future__ import annotations

import typing as t
import logging
import itertools
from typing import TYPE_CHECKING

logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from aiohttp import MultipartWriter
    from starlette.requests import Request

    from ..runner.container import Payload


T = t.TypeVar("T")
To = t.TypeVar("To")


CUDA_SUCCESS = 0


def pass_through(i: T) -> T:
    return i


class Params(t.Generic[T]):
    args: tuple[T, ...]
    kwargs: dict[str, T]

    def __init__(
        self,
        *args: T,
        **kwargs: T,
    ):
        self.args = args
        self.kwargs = kwargs

    def map(self, function: t.Callable[[T], To]) -> Params[To]:
        args = tuple(function(a) for a in self.args)
        kwargs = {k: function(v) for k, v in self.kwargs.items()}
        return Params[To](*args, **kwargs)

    def imap(self, function: t.Callable[[T], t.Iterable[To]]) -> t.Iterator[Params[To]]:
        args_iter = tuple(iter(function(a)) for a in self.args)
        kwargs_iter = {k: iter(function(v)) for k, v in self.kwargs.items()}

        try:
            while True:
                args = tuple(next(a) for a in args_iter)
                kwargs = {k: next(v) for k, v in kwargs_iter.items()}
                yield Params[To](*args, **kwargs)
        except StopIteration:
            pass

    def keys(self) -> t.Iterator[int | str]:
        return itertools.chain(range(len(self.args)), self.kwargs.keys())

    def items(self) -> t.Iterator[t.Tuple[t.Union[int, str], T]]:
        return itertools.chain(enumerate(self.args), self.kwargs.items())

    @classmethod
    def from_dict(cls, data: dict[str | int, T]) -> Params[T]:
        return cls(
            *(data[k] for k in sorted(k for k in data if isinstance(k, int))),
            **{k: v for k, v in data.items() if isinstance(k, str)},
        )

    @property
    def sample(self) -> T:
        if self.args:
            return self.args[0]
        return next(iter(self.kwargs.values()))


PAYLOAD_META_HEADER = "Bento-Payload-Meta"


def payload_params_to_multipart(params: Params[Payload]) -> MultipartWriter:
    import json

    from multidict import CIMultiDict
    from aiohttp.multipart import MultipartWriter

    multipart = MultipartWriter(subtype="form-data")
    for key, payload in params.items():
        multipart.append(
            payload.data,
            headers=CIMultiDict(
                (
                    (PAYLOAD_META_HEADER, json.dumps(payload.meta)),
                    ("Content-Type", f"application/vnd.bentoml.{payload.container}"),
                    ("Content-Disposition", f'form-data; name="{key}"'),
                )
            ),
        )
    return multipart


async def multipart_to_payload_params(request: Request) -> Params[Payload]:
    import json

    from bentoml._internal.runner.container import Payload
    from bentoml._internal.utils.formparser import populate_multipart_requests

    parts = await populate_multipart_requests(request)
    max_arg_index = -1
    kwargs: t.Dict[str, Payload] = {}
    args_map: t.Dict[int, Payload] = {}
    for field_name, req in parts.items():
        payload = Payload(
            data=await req.body(),
            meta=json.loads(req.headers[PAYLOAD_META_HEADER]),
            container=req.headers["Content-Type"].strip("application/vnd.bentoml."),
        )
        if field_name.isdigit():
            arg_index = int(field_name)
            args_map[arg_index] = payload
            max_arg_index = max(max_arg_index, arg_index)
        else:
            kwargs[field_name] = payload
    args = tuple(args_map[i] for i in range(max_arg_index + 1))
    return Params(*args, **kwargs)



from __future__ import annotations

import json
import typing as t
import logging
from typing import TYPE_CHECKING

from starlette.requests import Request
from starlette.responses import Response

from .base import IODescriptor
from .json import MIME_TYPE_JSON
from ..types import LazyType
from ..utils.http import set_cookies
from ...exceptions import BadInput
from ...exceptions import InternalServerError

if TYPE_CHECKING:
    import numpy as np

    from .. import external_typing as ext
    from ..context import InferenceApiContext as Context

logger = logging.getLogger(__name__)


def _is_matched_shape(
    left: t.Optional[t.Tuple[int, ...]],
    right: t.Optional[t.Tuple[int, ...]],
) -> bool:  # pragma: no cover
    if (left is None) or (right is None):
        return False

    if len(left) != len(right):
        return False

    for i, j in zip(left, right):
        if i == -1 or j == -1:
            continue
        if i == j:
            continue
        return False
    return True


class NumpyNdarray(IODescriptor["ext.NpNDArray"]):
    """
    :code:`NumpyNdarray` defines API specification for the inputs/outputs of a Service, where
    either inputs will be converted to or outputs will be converted from type
    :code:`numpy.ndarray` as specified in your API function signature.

    Sample implementation of a sklearn service:

    .. code-block:: python

        # sklearn_svc.py
        import bentoml
        from bentoml.io import NumpyNdarray
        import bentoml.sklearn

        runner = bentoml.sklearn.load_runner("sklearn_model_clf")

        svc = bentoml.Service("iris-classifier", runners=[runner])

        @svc.api(input=NumpyNdarray(), output=NumpyNdarray())
        def predict(input_arr):
            res = runner.run(input_arr)
            return res

    Users then can then serve this service with :code:`bentoml serve`:

    .. code-block:: bash

        % bentoml serve ./sklearn_svc.py:svc --auto-reload

        (Press CTRL+C to quit)
        [INFO] Starting BentoML API server in development mode with auto-reload enabled
        [INFO] Serving BentoML Service "iris-classifier" defined in "sklearn_svc.py"
        [INFO] API Server running on http://0.0.0.0:3000

    Users can then send requests to the newly started services with any client:

    .. tabs::

    .. code-block:: bash

        % curl -X POST -H "Content-Type: application/json" --data '[[5,4,3,2]]' http://0.0.0.0:3000/predict

        [1]%

    Args:
        dtype (:code:`numpy.typings.DTypeLike`, `optional`, default to :code:`None`):
            Data Type users wish to convert their inputs/outputs to. Refers to `arrays dtypes <https://numpy.org/doc/stable/reference/arrays.dtypes.html>`_ for more information.
        enforce_dtype (:code:`bool`, `optional`, default to :code:`False`):
            Whether to enforce a certain data type. if :code:`enforce_dtype=True` then :code:`dtype` must be specified.
        shape (:code:`Tuple[int, ...]`, `optional`, default to :code:`None`):
            Given shape that an array will be converted to. For example:

            .. code-block:: python

                from bentoml.io import NumpyNdarray

                @svc.api(input=inp(shape=(3,1), enforce_shape=True), output=NumpyNdarray())
                def predict(input_array: np.ndarray) -> np.ndarray:
                    # input_array will have shape (3,1)
                    result = await runner.run(input_array)

        enforce_shape (:code:`bool`, `optional`, default to :code:`False`):
            Whether to enforce a certain shape. If `enforce_shape=True` then `shape`
            must be specified

    Returns:
        :obj:`~bentoml._internal.io_descriptors.IODescriptor`: IO Descriptor that :code:`np.ndarray`.
    """

    def __init__(
        self,
        dtype: t.Optional[t.Union[str, "np.dtype[t.Any]"]] = None,
        enforce_dtype: bool = False,
        shape: t.Optional[t.Tuple[int, ...]] = None,
        enforce_shape: bool = False,
    ):
        import numpy as np

        if isinstance(dtype, str):
            dtype = np.dtype(dtype)

        self._dtype = dtype
        self._shape = shape
        self._enforce_dtype = enforce_dtype
        self._enforce_shape = enforce_shape

    def _infer_types(self) -> str:  # pragma: no cover
        if self._dtype is not None:
            name = self._dtype.name
            if name.startswith("int") or name.startswith("uint"):
                var_type = "integer"
            elif name.startswith("float") or name.startswith("complex"):
                var_type = "number"
            else:
                var_type = "object"
        else:
            var_type = "object"
        return var_type

    def _items_schema(self) -> t.Dict[str, t.Any]:
        if self._shape is not None:
            if len(self._shape) > 1:
                return {"type": "array", "items": {"type": self._infer_types()}}
            return {"type": self._infer_types()}
        return {}

    def input_type(self) -> LazyType["ext.NpNDArray"]:
        return LazyType("numpy", "ndarray")

    def openapi_schema_type(self) -> t.Dict[str, t.Any]:
        return {"type": "array", "items": self._items_schema()}

    def openapi_request_schema(self) -> t.Dict[str, t.Any]:
        """Returns OpenAPI schema for incoming requests"""
        return {MIME_TYPE_JSON: {"schema": self.openapi_schema_type()}}

    def openapi_responses_schema(self) -> t.Dict[str, t.Any]:
        """Returns OpenAPI schema for outcoming responses"""
        return {MIME_TYPE_JSON: {"schema": self.openapi_schema_type()}}

    def _verify_ndarray(
        self,
        obj: "ext.NpNDArray",
        exception_cls: t.Type[Exception] = BadInput,
    ) -> "ext.NpNDArray":
        if self._dtype is not None and self._dtype != obj.dtype:
            if self._enforce_dtype:
                raise exception_cls(
                    f"{self.__class__.__name__}: enforced dtype mismatch"
                )
            try:
                obj = obj.astype(self._dtype)  # type: ignore
            except ValueError as e:
                logger.warning(f"{self.__class__.__name__}: {e}")

        if self._shape is not None and not _is_matched_shape(self._shape, obj.shape):
            if self._enforce_shape:
                raise exception_cls(
                    f"{self.__class__.__name__}: enforced shape mismatch"
                )
            try:
                obj = obj.reshape(self._shape)
            except ValueError as e:
                logger.warning(f"{self.__class__.__name__}: {e}")
        return obj

    async def from_http_request(self, request: Request) -> "ext.NpNDArray":
        """
        Process incoming requests and convert incoming
         objects to `numpy.ndarray`

        Args:
            request (`starlette.requests.Requests`):
                Incoming Requests
        Returns:
            a `numpy.ndarray` object. This can then be used
             inside users defined logics.
        """
        import numpy as np

        obj = await request.json()
        res: "ext.NpNDArray"
        try:
            res = np.array(obj, dtype=self._dtype)  # type: ignore[arg-type]
        except ValueError:
            res = np.array(obj)  # type: ignore[arg-type]
        res = self._verify_ndarray(res, BadInput)
        return res

    async def to_http_response(self, obj: ext.NpNDArray, ctx: Context | None = None):
        """
        Process given objects and convert it to HTTP response.

        Args:
            obj (`np.ndarray`):
                `np.ndarray` that will be serialized to JSON
        Returns:
            HTTP Response of type `starlette.responses.Response`. This can
             be accessed via cURL or any external web traffic.
        """
        obj = self._verify_ndarray(obj, InternalServerError)
        if ctx is not None:
            res = Response(
                json.dumps(obj.tolist()),
                media_type=MIME_TYPE_JSON,
                headers=ctx.response.metadata,  # type: ignore (bad starlette types)
                status_code=ctx.response.status_code,
            )
            set_cookies(res, ctx.response.cookies)
            return res
        else:
            return Response(json.dumps(obj.tolist()), media_type=MIME_TYPE_JSON)

    @classmethod
    def from_sample(
        cls,
        sample_input: "ext.NpNDArray",
        enforce_dtype: bool = True,
        enforce_shape: bool = True,
    ) -> "NumpyNdarray":
        """
        Create a NumpyNdarray IO Descriptor from given inputs.

        Args:
            sample_input (:code:`np.ndarray`): Given sample np.ndarray data
            enforce_dtype (;code:`bool`, `optional`, default to :code:`True`):
                Enforce a certain data type. :code:`dtype` must be specified at function
                signature. If you don't want to enforce a specific dtype then change
                :code:`enforce_dtype=False`.
            enforce_shape (:code:`bool`, `optional`, default to :code:`False`):
                Enforce a certain shape. :code:`shape` must be specified at function
                signature. If you don't want to enforce a specific shape then change
                :code:`enforce_shape=False`.

        Returns:
            :obj:`~bentoml._internal.io_descriptors.NumpyNdarray`: :code:`NumpyNdarray` IODescriptor from given users inputs.

        Example:

        .. code-block:: python

            import numpy as np
            from bentoml.io import NumpyNdarray
            arr = [[1,2,3]]
            inp = NumpyNdarray.from_sample(arr)

            ...
            @svc.api(input=inp, output=NumpyNdarray())
            def predict() -> np.ndarray:...
        """
        return cls(
            dtype=sample_input.dtype,
            shape=sample_input.shape,
            enforce_dtype=enforce_dtype,
            enforce_shape=enforce_shape,
        )



import typing as t
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from . import Service
    from ..io_descriptors import IODescriptor


HEALTHZ_DESC = (
    "Health check endpoint. Expecting an empty response with status code"
    " <code>200</code> when the service is in health state. The <code>/healthz</code>"
    " endpoint is <b>deprecated</b> (since Kubernetes v1.16)"
)

LIVEZ_DESC = (
    "Health check endpoint for Kubernetes. Healthy endpoint responses with a"
    " <code>200</code> OK status."
)

READYZ_DESC = (
    "A <code>200</code> OK status from <code>/readyz</code> endpoint indicated"
    " the service is ready to accept traffic. From that point and onward, Kubernetes"
    " will use <code>/livez</code> endpoint to perform periodic health checks."
)

METRICS_DESC = "Prometheus metrics endpoint"


def _generate_responses_schema(
    output: "IODescriptor[t.Any]",
) -> t.Dict[str, t.Dict[str, t.Any]]:
    resp = {
        "200": dict(
            description="success",
            content=output.openapi_responses_schema(),
        ),
        "400": dict(description="Bad Request"),
        "404": dict(description="Not Found"),
        "500": dict(description="Internal Server Error"),
    }
    return resp


def get_service_openapi_doc(svc: "Service"):
    info = {
        "title": svc.name,
        "description": "A Prediction Service built with BentoML",
        "contact": {"email": "contact@bentoml.ai"},
        "version": svc.tag.version if svc.tag is not None else "0.0.0",
    }
    docs: t.Dict[str, t.Any] = {
        "openapi": "3.0.0",
        "info": info,
        "tags": [
            {
                "name": "infra",
                "description": "Infrastructure endpoints",
            },
            {
                "name": "app",
                "description": "Inference endpoints",
            },
        ],
    }

    paths = {}
    default_response = {"200": {"description": "success"}}

    paths["/healthz"] = {
        "get": {
            "tags": ["infra"],
            "description": HEALTHZ_DESC,
            "responses": default_response,
        }
    }
    paths["/livez"] = {
        "get": {
            "tags": ["infra"],
            "description": LIVEZ_DESC,
            "responses": default_response,
        }
    }
    paths["/readyz"] = {
        "get": {
            "tags": ["infra"],
            "description": READYZ_DESC,
            "responses": default_response,
        }
    }
    paths["/metrics"] = {
        "get": {
            "tags": ["infra"],
            "description": METRICS_DESC,
            "responses": default_response,
        }
    }

    for api in svc.apis.values():
        api_path = api.route if api.route.startswith("/") else f"/{api.route}"

        paths[api_path] = {
            "post": dict(
                tags=["app"],
                summary=f"{api}",
                description=api.doc or "",
                operationId=f"{svc.name}__{api.name}",
                requestBody=dict(content=api.input.openapi_request_schema()),
                responses=_generate_responses_schema(api.output),
                # examples=None,
                # headers=None,
            )
        }

    docs["paths"] = paths
    return docs



import os
import re
import typing as t
import logging
from typing import TYPE_CHECKING
from distutils.dir_util import copy_tree

from simple_di import inject
from simple_di import Provide

import bentoml
from bentoml import Tag
from bentoml.exceptions import NotFound
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..utils import LazyLoader
from ..models import SAVE_NAMESPACE
from ..utils.pkg import get_pkg_version
from .common.model_runner import BaseModelRunner
from ..configuration.containers import BentoMLContainer

logger = logging.getLogger(__name__)

MODULE_NAME = "bentoml.paddle"

PADDLE_MODEL_EXTENSION = ".pdmodel"
PADDLE_PARAMS_EXTENSION = ".pdiparams"

_paddle_exc = """\
Instruction for installing `paddlepaddle`:
- CPU support only: `pip install paddlepaddle -i https://mirror.baidu.com/pypi/simple`
- GPU support: (latest version): `python -m pip install paddlepaddle-gpu==2.1.3.post112 -f https://www.paddlepaddle.org.cn/whl/linux/mkl/avx/stable.html`.
    For other version of CUDA or different platforms refer to https://www.paddlepaddle.org.cn/ for more information
"""

if TYPE_CHECKING:
    import numpy as np
    import paddle
    import paddle.nn
    import paddlehub as hub
    import paddle.inference
    import paddlehub.module.module as module
    from paddle.static import InputSpec
    from paddle.fluid.dygraph.dygraph_to_static.program_translator import StaticFunction

    from ..models import ModelStore

try:
    import paddle
    import paddle.nn
    import paddle.inference
    from paddle.fluid import core
except ImportError:  # pragma: no cover
    raise MissingDependencyException(_paddle_exc)

_hub_exc = (
    """\
`paddlehub` is required to use `bentoml.paddle.import_from_paddlehub()`. Make sure
 have `paddlepaddle` installed beforehand. Install `paddlehub` with
 `pip install paddlehub`.
"""
    + _paddle_exc
)

# TODO: supports for PIL.Image and pd.DataFrame?
# try:
#     import PIL
#     import PIL.Image
# except ImportError:
#     PIL = None
#     PIL.Image = None
# _PIL_warning = """\
# `Pillow` is optionally required to use `bentoml.paddle._PaddlePaddleRunner._run_batch`.
# Instruction: `pip install -U Pillow`
# """
#
# try:
#     import pandas as pd
# except ImportError:
#     pd = None
# _pd_warning = """\
# `pandas` is optionally required to use `bentoml.paddle._PaddlePaddleRunner._run_batch`.
# Instruction: `pip install -U pandas`
# """

hub = LazyLoader("hub", globals(), "paddlehub", exc_msg=_hub_exc)
manager = LazyLoader("manager", globals(), "paddlehub.module.manager", exc_msg=_hub_exc)
server = LazyLoader("server", globals(), "paddlehub.server.server", exc_msg=_hub_exc)
np = LazyLoader("np", globals(), "numpy")


def _clean_name(name: str) -> str:  # pragma: no cover
    return re.sub(r"\W|^(?=\d)-", "_", name)


@inject
def _load_paddle_bentoml_default_config(
    tag: t.Union[str, Tag],
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "paddle.inference.Config":
    # https://github.com/PaddlePaddle/Paddle/blob/develop/paddle/fluid/inference/api/analysis_config.cc
    model = model_store.get(tag)
    config = paddle.inference.Config(
        model.path_of(f"{SAVE_NAMESPACE}{PADDLE_MODEL_EXTENSION}"),
        model.path_of(f"{SAVE_NAMESPACE}{PADDLE_PARAMS_EXTENSION}"),
    )
    config.enable_memory_optim()
    config.switch_ir_optim(True)
    return config


@inject
def load(
    tag: t.Union[str, Tag],
    config: t.Optional["paddle.inference.Config"] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
    **kwargs: str,
) -> t.Union[
    "paddle.inference.Predictor", t.Union["module.RunModule", "module.ModuleV1"]
]:
    """
    Load a model from BentoML local modelstore with given name.

    Args:
        tag (:code:`Union[str, Tag]`):
            Tag of a saved model in BentoML local modelstore.
        config (`paddle.inference.Config`, `optional`, default to :code:`None`):
            Model config to be used to create a `Predictor` instance.
        model_store (:mod:`~bentoml._internal.models.store.ModelStore`, default to :mod:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj::code:`Union[paddle.inference.Predictor, paddlehub.module.module.RunModule, paddlehub.module.module.ModuleV1]`: an instance of
        one of :obj:`paddle.inference.Predictor`, :obj:`paddlehub.module.module.RunModule`, and :obj:`paddlehub.module.module.ModuleV1`
        from BentoML modelstore.

    Examples:

    .. code-block:: python

        import bentoml

        model = bentoml.paddle.load(tag)

    """
    model = model_store.get(tag)
    if model.info.module not in (MODULE_NAME, __name__):
        raise BentoMLException(
            f"Model {tag} was saved with module {model.info.module}, failed loading with {MODULE_NAME}."
        )
    if "paddlehub" in model.info.context:
        if model.info.options["from_local_dir"]:
            return hub.Module(directory=model.path)
        else:
            server.CacheUpdater(
                "update_cache",
                module=model.info.options["name"],
                version=model.info.options["version"],
            ).start()
            directory = (
                model.path
                if "_module_dir" not in model.info.options
                else model.path_of(model.info.options["_module_dir"])
            )
            return hub.Module(directory=directory, **kwargs)
    else:
        _config = _load_paddle_bentoml_default_config(tag) if not config else config
        return paddle.inference.create_predictor(_config)


def _save(
    name: str,
    model: t.Union[str, "paddle.nn.Layer", "paddle.inference.Predictor"],
    version: t.Optional[str],
    branch: t.Optional[str],
    source: t.Optional[str],
    update: bool,
    ignore_env_mismatch: bool,
    hub_module_home: t.Optional[str],
    keep_download_from_hub: bool,
    *,
    input_spec: t.Optional[t.Union[t.List["InputSpec"], t.Tuple["InputSpec", ...]]],
    labels: t.Optional[t.Dict[str, str]] = None,
    custom_objects: t.Optional[t.Dict[str, t.Any]] = None,
    metadata: t.Optional[t.Dict[str, t.Any]],
    model_store: "ModelStore",
) -> Tag:
    context: t.Dict[str, t.Any] = {
        "framework_name": "paddle",
        "pip_dependencies": [f"paddlepaddle=={get_pkg_version('paddlepaddle')}"],
    }
    if isinstance(model, str):
        context["paddlehub"] = hub.__version__
        if not os.path.isdir(model):
            try:  # pragma: no cover
                # NOTE: currently there is no need to test this feature.
                _model = model_store.get(model)
                if not keep_download_from_hub:
                    logger.warning(
                        f"""\
`{name}` is found under BentoML modelstore. Returning {_model.tag}...

Mechanism: `bentoml.paddle.import_from_paddlehub` will initialize an instance of
 `LocalModuleManager` from `paddlehub`. We will check whether the module is already
 downloaded under `paddlehub` cache, then copy over to BentoML modelstore.
The reason behind the design is due to internal mechanism of `hub.Module` with regarding
 inference. If we save the module in Paddle inference format, there is no guarantee that
 users can use Module functionality. Refers to https://paddlehub.readthedocs.io/en/release-v2.1/tutorial/custom_module.html
 for how a Module is structured.

For most use-case where you are using pretrained model provided by PaddlePaddle
 (https://www.paddlepaddle.org.cn/hublist), since {_model.tag} exists, this means you
 have previously imported the module from paddlehub to BentoML modelstore, we will
 return the existing tag.
For use-case where you have a custom `hub.Module` or wanting to use different iteration
 of the aforementioned pretrained model, specify `keep_download_from_hub=True`,
 `version=<your_specific_version>` or any other related `kwargs`. Refers to
 https://paddlehub.readthedocs.io/en/release-v2.1/api/module.html for more information.
                        """
                    )
                    _tag = _model.tag
                    server.CacheUpdater(
                        "update_cache", module=model, version=version
                    ).start()
                    return _tag
            except (FileNotFoundError, NotFound):
                pass

    with bentoml.models.create(
        name,
        module=MODULE_NAME,
        context=context,
        labels=labels,
        custom_objects=custom_objects,
        metadata=metadata,
    ) as _model:
        if isinstance(model, str):
            # NOTE: for paddlehub there is no way to skip Module initialization,
            #  since `paddlehub` will always initialize a `hub.Module` regardless
            #  of any situation. Therefore, the bottleneck will only happen one time
            #  when users haven't saved the pretrained model under paddlehub cache
            #  directory.
            if not hub.server_check():
                raise BentoMLException("Unable to connect to PaddleHub server.")
            if os.path.isdir(model):
                directory = model
                target = _model.path

                _model.info.options["from_local_dir"] = True
            else:
                _local_manager = manager.LocalModuleManager(home=hub_module_home)
                user_module_cls = _local_manager.search(
                    name, source=source, branch=branch
                )
                if not user_module_cls or not user_module_cls.version.match(version):
                    user_module_cls = _local_manager.install(
                        name=name,
                        version=version,
                        source=source,
                        update=update,
                        branch=branch,
                        ignore_env_mismatch=ignore_env_mismatch,
                    )

                directory = _local_manager._get_normalized_path(user_module_cls.name)
                target = _model.path_of(user_module_cls.name)
                print(target)

                _model.info.options = {}
                _model.info.options.update(hub.Module.load_module_info(directory))
                _model.info.options["_module_dir"] = os.path.relpath(
                    target, _model.path
                )
                _model.info.options["from_local_dir"] = False
            copy_tree(directory, target)
        else:
            paddle.jit.save(
                model, _model.path_of(SAVE_NAMESPACE), input_spec=input_spec
            )

        return _model.tag


@inject
def save(
    name: str,
    model: t.Union["paddle.nn.Layer", "paddle.inference.Predictor", "StaticFunction"],
    *,
    input_spec: t.Optional[
        t.Union[t.List["InputSpec"], t.Tuple["InputSpec", ...]]
    ] = None,
    labels: t.Optional[t.Dict[str, str]] = None,
    custom_objects: t.Optional[t.Dict[str, t.Any]] = None,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (:code:`str`):
            Name for given model instance. This should pass Python identifier check.
        model (:code:`Union[paddle.nn.Layer, paddle.inference.Predictor, StaticFunction]`):
            Instance of :code:`paddle.nn.Layer`, decorated functions, or
            :code:`paddle.inference.Predictor` to be saved.
        input_spec (:code:`Union[List[InputSpec], Tuple[InputSpec, ...]]`, `optional`, default to :code:`None`):
            Describes the input of the saved model's forward method, which can be
            described by :obj:`InputSpec` or example Tensor. Moreover, we support to specify
            non-tensor type argument, such as `int`, `float`, `string`, or
            `list`/`dict` of them. If `None`, all input variables of the original
            Layer's forward method would be the inputs of the saved model. Generally
            this is **NOT RECOMMENDED** to use unless you know what you are doing.
        labels (:code:`Dict[str, str]`, `optional`, default to :code:`None`):
            user-defined labels for managing models, e.g. team=nlp, stage=dev
        custom_objects (:code:`Dict[str, Any]]`, `optional`, default to :code:`None`):
            user-defined additional python objects to be saved alongside the model,
            e.g. a tokenizer instance, preprocessor function, model configuration json
        metadata (:code:`Dict[str, Any]`, `optional`,  default to :code:`None`):
            Custom metadata for given model.
        model_store (:mod:`~bentoml._internal.models.store.ModelStore`, default to :mod:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format `name:version` where `name` is the user-defined model's name, and a generated `version` by BentoML.

    Examples:

    .. code-block:: python

        import random
        import numpy as np

        import bentoml
        import paddle
        import paddle.nn as nn
        from paddle.static import InputSpec

        IN_FEATURES = 13
        OUT_FEATURES = 1


        def set_random_seed(seed):
            random.seed(seed)
            np.random.seed(seed)
            paddle.seed(seed)
            paddle.framework.random._manual_program_seed(seed)


        class LinearModel(nn.Layer):
            def __init__(self):
                super(LinearModel, self).__init__()
                self.fc = nn.Linear(IN_FEATURES, OUT_FEATURES)

            @paddle.jit.to_static(input_spec=[InputSpec(shape=[IN_FEATURES], dtype="float32")])
            def forward(self, x):
                return self.fc(x)


        def train_paddle_model() -> "LinearModel":
            set_random_seed(SEED)
            model = LinearModel()
            loss = nn.MSELoss()
            adam = paddle.optimizer.Adam(parameters=model.parameters())

            train_data = paddle.text.datasets.UCIHousing(mode="train")

            loader = paddle.io.DataLoader(
                train_data, batch_size=BATCH_SIZE, shuffle=True, drop_last=True, num_workers=2
            )

            model.train()
            for _ in range(EPOCH_NUM):
                for _, (feature, label) in enumerate(loader()):
                    out = model(feature)
                    loss_fn = loss(out, label)
                    loss_fn.backward()
                    adam.step()
                    adam.clear_grad()
            return model

        model = train_paddle_model()
        # `save` a pretrained model to BentoML modelstore:
        tag = bentoml.paddle.save("linear_model", model, input_spec=InputSpec(shape=[IN_FEATURES], dtype="float32"))
    """
    return _save(
        name=name,
        model=model,
        input_spec=input_spec,
        labels=labels,
        custom_objects=custom_objects,
        metadata=metadata,
        model_store=model_store,
        version=None,
        branch=None,
        source=None,
        update=False,
        hub_module_home=None,
        ignore_env_mismatch=False,
        keep_download_from_hub=False,
    )


@inject
def import_from_paddlehub(
    model_name: str,
    name: t.Optional[str] = None,
    version: t.Optional[str] = None,
    branch: t.Optional[str] = None,
    source: t.Optional[str] = None,
    update: bool = False,
    ignore_env_mismatch: bool = False,
    hub_module_home: t.Optional[str] = None,
    keep_download_from_hub: bool = False,
    *,
    labels: t.Optional[t.Dict[str, str]] = None,
    custom_objects: t.Optional[t.Dict[str, t.Any]] = None,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> Tag:
    """
    Import models from :code:`PaddleHub` and save it under BentoML modelstore.

    Args:
        model_name (:code:`str`):
            Name for a PaddleHub model. This can be either path to a Hub module or model
            name, for both v2 and v1.
        name (:code:`str`, `optional`, default to :code:`None`):
            Name for given model instance. This should pass Python identifier check. If
            not specified, then BentoML will save under :obj:`model_name`.
        version (:code:`str`, `optional`, default to :code:`None`):
            The version limit of the module, only takes effect when the :obj:`name` is
            specified. When the local Module does not meet the specified version
            conditions, PaddleHub will re-request the server to download the
            appropriate Module. Default to `None`, This means that the local Module will
            be used. If the Module does not exist, PaddleHub will download the latest
            version available from the server according to the usage environment.
        source (:code:`str`, `optional`, default to :code:`None`):
            URL of a git repository. If this parameter is specified, PaddleHub will no
            longer download the specified Module from the default server, but will look
            for it in the specified repository.
        update (`bool`, `optional`, default to :code:`False`):
            Whether to update the locally cached git repository, only takes effect when
            the `source` is specified.
        branch (:code:`str`, `optional`, default to :code:`None`):
            The branch of the specified git repository.
        ignore_env_mismatch (`bool`, `optional`, default to :code:`False`):
            Whether to ignore the environment mismatch when installing the Module.
        hub_module_home (:code:`str`, `optional`, default is `None`):
            Location to save for your PaddleHub cache. If `None`, then default to use
            PaddleHub default cache, which is under :code:`$HOME/.paddlehub`
        keep_download_from_hub (`bool`, `optional`, default to :code:`False`):
            Whether to re-download pretrained model from hub.
        labels (:code:`Dict[str, str]`, `optional`, default to :code:`None`):
            user-defined labels for managing models, e.g. team=nlp, stage=dev
        custom_objects (:code:`Dict[str, Any]]`, `optional`, default to :code:`None`):
            user-defined additional python objects to be saved alongside the model,
            e.g. a tokenizer instance, preprocessor function, model configuration json
        metadata (:code:`Dict[str, Any]`, `optional`,  default to :code:`None`):
            Custom metadata for given model.
        model_store (:mod:`~bentoml._internal.models.store.ModelStore`, default to :mod:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format `name:version` where `name` is the user-defined model's name, and a generated `version` by BentoML.

    Examples:

    .. code-block:: python

        import bentoml

        tag = bentoml.paddle.import_from_paddlehub("senta_bilstm")

    """
    _name = model_name.split("/")[-1] if os.path.isdir(model_name) else model_name

    return _save(
        name=_clean_name(_name) if not name else name,
        model=model_name,
        input_spec=None,
        version=version,
        branch=branch,
        source=source,
        update=update,
        ignore_env_mismatch=ignore_env_mismatch,
        hub_module_home=hub_module_home,
        keep_download_from_hub=keep_download_from_hub,
        labels=labels,
        custom_objects=custom_objects,
        metadata=metadata,
        model_store=model_store,
    )


class _PaddlePaddleRunner(BaseModelRunner):
    def __init__(
        self,
        tag: t.Union[str, Tag],
        infer_api_callback: str,
        *,
        gpu_mem_pool_mb: int = 1024,
        config: t.Optional["paddle.inference.Config"],
        name: t.Optional[str] = None,
    ):
        super().__init__(tag=tag, name=name)

        self._infer_api_callback = infer_api_callback
        self._config = config
        self._gpu_mem_pool_mb = gpu_mem_pool_mb

    @property
    def _enable_gpu(self):
        enable_gpu = self.resource_quota.nvidia_gpu
        if enable_gpu and not paddle.is_compiled_with_cuda():  # type: ignore
            raise BentoMLException(
                "`resource_quota.nvidia_com_gpu` is not 0 while CUDA is not currently supported by existing paddlepaddle."
                " Make sure to install `paddlepaddle-gpu` and try again."
            )
        return enable_gpu

    def _setup_runner_config(self) -> None:
        _config = (
            _load_paddle_bentoml_default_config(self._tag)
            if not self._config
            else self._config
        )
        if self._enable_gpu:
            _config.enable_use_gpu(self._gpu_mem_pool_mb)
        else:
            # If not specific mkldnn, you can set the blas thread.
            # `num_threads` should not be greater than the number of cores in the CPU.
            _config.set_cpu_math_library_num_threads(self._num_threads)
            _config.enable_mkldnn()
            _config.disable_gpu()
        paddle.set_device(self._device)
        self._runner_config = _config

    @property
    def _device(self) -> str:
        if self._enable_gpu:
            # this will default to gpu:0
            return "gpu"
        return "cpu"

    @property
    def _num_threads(self) -> int:
        if self._enable_gpu and self.resource_quota.nvidia_gpu:
            return 1
        return max(round(self.resource_quota.cpu), 1)

    @property
    def num_replica(self) -> int:
        if self._enable_gpu:
            return self.resource_quota.nvidia_gpu
        return 1

    def _setup(self) -> None:
        self._setup_runner_config()
        self._model = load(self._tag, config=self._config, model_store=self.model_store)
        self._infer_func = getattr(self._model, self._infer_api_callback)

    def _run_batch(
        self,
        input_data: t.Optional[t.Any],
        *args: str,
        return_argmax: bool = False,
        **kwargs: str,
    ) -> t.Any:
        model_info = self.model_info
        if "paddlehub" in model_info.info.context:
            return self._infer_func(*args, **kwargs)
        else:
            assert input_data is not None
            res = list()
            if isinstance(input_data, paddle.Tensor):
                input_data = input_data.numpy()
            # TODO: supports for PIL.Image and DataFrame?
            # else:
            #     if PIL is not None:
            #         if isinstance(input_data, PIL.Image.Image):
            #             input_data = input_data.fromarray(input_data)
            #     else:
            #         logger.warning(_PIL_warning)
            #     if pd is not None:
            #         if isinstance(input_data, pd.DataFrame):
            #             input_data = input_data.to_numpy(dtype=np.float32)
            #         else:
            #             logger.warning(_pd_warning)

            input_names = self._model.get_input_names()

            for i, name in enumerate(input_names):
                input_tensor = self._model.get_input_handle(name)
                input_tensor.reshape(input_data[i].shape)
                input_tensor.copy_from_cpu(input_data[i].copy())

            self._infer_func()

            output_names = self._model.get_output_names()
            for i, name in enumerate(output_names):
                output_tensor = self._model.get_output_handle(name)
                output_data = output_tensor.copy_to_cpu()
                res.append(np.asarray(output_data))
            if return_argmax:
                return np.argmax(res, dims=-1)  # type: ignore[call-overload]
            return res


@inject
def load_runner(
    tag: t.Union[str, Tag],
    *,
    infer_api_callback: str = "run",
    gpu_mem_pool_mb: int = 0,
    name: t.Optional[str] = None,
    config: t.Optional["paddle.inference.Config"] = None,
) -> "_PaddlePaddleRunner":
    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. :func:`bentoml.paddle.load_runner` implements a Runner class that
    either wraps around :obj:`paddle.inference.Predictor` or :obj:`paddlehub.module.Module`,
    which optimize it for the BentoML runtime.

    Args:
        tag (:code:`Union[str, Tag]`):
            Tag of a saved model in BentoML local modelstore.
        infer_api_callback (:code:`str`, `optional`, default to :code:`run`):
            Inference API function that will be used during :obj:`run_batch`. If :obj:`tag`
            is a tag of a :obj:`hub.Module`, then :code:`infer_api_callback` should be changed
            to corresponding API endpoint offered by given module.
        config (`paddle.inference.Config`, `optional`, default to :code:`None`):
            Config for inference. If None is specified, then use BentoML default.
        gpu_mem_pool_mb (`int`, `optional`, default to 0):
            Amount of memory one wants to allocate to GPUs. By default we will allocate None.

    Returns:
        :obj:`~bentoml._internal.runner.Runner`: Runner instances for :mod:`bentoml.paddle` model

    Examples:

    .. code-block:: python

        import bentoml
        import numpy as np

        runner = bentoml.paddle.load_runner(
            tag,
            device="gpu:0",
        )

        _ = runner.run_batch(pd_dataframe.to_numpy().astype(np.float32))

    """
    return _PaddlePaddleRunner(
        tag=tag,
        infer_api_callback=infer_api_callback,
        gpu_mem_pool_mb=gpu_mem_pool_mb,
        config=config,
        name=name,
    )



import signal
import subprocess
from typing import TYPE_CHECKING

import psutil

if TYPE_CHECKING:
    import typing as t


def kill_subprocess_tree(p: "subprocess.Popen[t.Any]") -> None:
    """
    Tell the process to terminate and kill all of its children. Availabe both on Windows and Linux.
    Note: It will return immediately rather than wait for the process to terminate.

    Args:
        p: subprocess.Popen object
    """
    if psutil.WINDOWS:
        subprocess.call(["taskkill", "/F", "/T", "/PID", str(p.pid)])
    else:
        p.terminate()


def cancel_subprocess(p: "subprocess.Popen[t.Any]") -> None:
    if psutil.WINDOWS:
        p.send_signal(signal.CTRL_C_EVENT)  # type: ignore
    else:
        p.send_signal(signal.SIGINT)



from __future__ import annotations

import os
import re
import math
import typing as t
import logging
import functools

import attr
import psutil
from cattrs.gen import override
from cattrs.gen import make_dict_unstructure_fn

from ..utils import bentoml_cattr
from ...exceptions import BentoMLException

logger = logging.getLogger(__name__)


@attr.define(frozen=True)
class Resource:
    cpu: t.Optional[float] = attr.field(default=None)
    nvidia_gpu: t.Optional[float] = attr.field(default=None)
    custom_resources: t.Dict[str, float] = attr.field(factory=dict)

    def __or__(self, right: Resource) -> Resource:
        """
        Fill in missing values with values from another.
        """
        cpu = right.cpu if self.cpu is None else self.cpu
        nvidia_gpu = right.nvidia_gpu if self.nvidia_gpu is None else self.nvidia_gpu
        custom_resources = dict(
            right.custom_resources,
            **{k: v for k, v in self.custom_resources.items() if v is not None},
        )
        return self.__class__(
            cpu=cpu,
            nvidia_gpu=nvidia_gpu,
            custom_resources=custom_resources,
        )

    @classmethod
    def from_config(cls, runner_name: str) -> Resource:
        """
        Create a Resource object from the BentoML config.
        """
        # TODO(jiang)
        return cls()

    @classmethod
    def from_system(cls) -> Resource:
        """
        Get Resource from system.
        """
        cpu = query_cpu_count()
        nvidia_gpu = float(query_nvidia_gpu_count())
        return cls(cpu=cpu, nvidia_gpu=nvidia_gpu)


# Remove after attrs support ForwardRef natively
attr.resolve_types(Resource, globals(), locals())

bentoml_cattr.register_unstructure_hook(
    Resource,
    make_dict_unstructure_fn(
        Resource,
        bentoml_cattr,
        cpu=override(omit_if_default=True),
        nvidia_gpu=override(omit_if_default=True),
        custom_resources=override(omit_if_default=True),
    ),
)


def cpu_converter(cpu: t.Union[int, float, str]) -> float:
    """
    Convert cpu to float.

    cpu can be a float, int or string.
    - 10m -> 0.01
    - 1.0 -> 1.0
    - 1 -> 1.0
    - "1" -> 1.0
    """
    assert isinstance(cpu, (int, float, str)), "cpu must be int, float or str"

    if isinstance(cpu, (int, float)):
        return float(cpu)

    milli_match = re.match("([0-9]+)m", cpu)
    if milli_match:
        return float(milli_match[1]) / 1000.0
    raise BentoMLException(f"Invalid CPU resource limit '{cpu}'. ")


def mem_converter(mem: t.Union[int, str]) -> int:
    assert isinstance(mem, (int, str)), "mem must be int or str"

    if isinstance(mem, int):
        return mem

    unit_match = re.match("([0-9]+)([A-Za-z]{1,2})", mem)
    mem_multipliers = {
        "k": 1000,
        "M": 1000**2,
        "G": 1000**3,
        "T": 1000**4,
        "P": 1000**5,
        "E": 1000**6,
        "Ki": 1024,
        "Mi": 1024**2,
        "Gi": 1024**3,
        "Ti": 1024**4,
        "Pi": 1024**5,
        "Ei": 1024**6,
    }
    if unit_match:
        base = int(unit_match[1])
        unit = unit_match[2]
        if unit in mem_multipliers:
            return base * mem_multipliers[unit]
    raise ValueError(f"Invalid MEM resource limit '{mem}'")


@functools.lru_cache(maxsize=1)
def query_cgroup_cpu_count() -> float:
    # Query active cpu processor count using cgroup v1 API, based on OpenJDK
    # implementation for `active_processor_count` using cgroup v1:
    # https://github.com/openjdk/jdk/blob/master/src/hotspot/os/linux/cgroupSubsystem_linux.cpp
    # For cgroup v2, see:
    # https://github.com/openjdk/jdk/blob/master/src/hotspot/os/linux/cgroupV2Subsystem_linux.cpp
    # Possible supports: cpuset.cpus on kubernetes
    def _read_cgroup_file(filename: str) -> float:
        with open(filename, "r", encoding="utf-8") as f:
            return int(f.read().strip())

    cgroup_root = "/sys/fs/cgroup/"
    cfs_quota_us_file = os.path.join(cgroup_root, "cpu", "cpu.cfs_quota_us")
    cfs_period_us_file = os.path.join(cgroup_root, "cpu", "cpu.cfs_period_us")
    shares_file = os.path.join(cgroup_root, "cpu", "cpu.shares")
    cpu_max_file = os.path.join(cgroup_root, "cpu.max")

    quota, shares = None, None

    if os.path.exists(cfs_quota_us_file) and os.path.exists(cfs_period_us_file):
        try:
            quota = _read_cgroup_file(cfs_quota_us_file) / _read_cgroup_file(
                cfs_period_us_file
            )
        except FileNotFoundError as err:
            logger.warning(f"Caught exception while calculating CPU quota: {err}")
    # reading from cpu.max for cgroup v2
    elif os.path.exists(cpu_max_file):
        try:
            with open(cpu_max_file, "r") as max_file:
                cfs_string = max_file.read()
                quota_str, period_str = cfs_string.split()
                if quota_str.isnumeric() and period_str.isnumeric():
                    quota = float(quota_str) / float(period_str)
                else:
                    # quota_str is "max" meaning the cpu quota is unset
                    quota = None
        except FileNotFoundError as err:
            logger.warning(f"Caught exception while calculating CPU quota: {err}")
    if quota is not None and quota < 0:
        quota = None
    elif quota == 0:
        quota = 1

    if os.path.exists(shares_file):
        try:
            shares = _read_cgroup_file(shares_file) / float(1024)
        except FileNotFoundError as err:
            logger.warning(f"Caught exception while getting CPU shares: {err}")

    os_cpu_count = float(os.cpu_count() or 1.0)

    limit_count = math.inf

    if quota and shares:
        limit_count = min(quota, shares)
    elif quota:
        limit_count = quota
    elif shares:
        limit_count = shares

    return float(min(limit_count, os_cpu_count))


@functools.lru_cache(maxsize=1)
def query_os_cpu_count() -> int:
    cpu_count = os.cpu_count()
    if cpu_count is not None:
        return cpu_count
    logger.warning("os.cpu_count() is NoneType")
    return 1


def query_cpu_count() -> float:
    # Default to the total CPU count available in current node or cgroup
    if psutil.POSIX:
        return query_cgroup_cpu_count()
    else:
        return float(query_os_cpu_count())


@functools.lru_cache(maxsize=1)
def query_nvidia_gpu_count() -> int:
    """
    query nvidia gpu count, available on Windows and Linux
    """
    import pynvml  # type: ignore

    try:
        pynvml.nvmlInit()
        device_count = pynvml.nvmlDeviceGetCount()
        return device_count
    except (pynvml.nvml.NVMLError, OSError):
        return 0
    finally:
        try:
            pynvml.nvmlShutdown()
        except Exception:  # pylint: disable=broad-except
            pass


def gpu_converter(gpus: t.Optional[t.Union[int, str, t.List[str]]]) -> int:
    if isinstance(gpus, int):
        return gpus

    if isinstance(gpus, str):
        return int(gpus)

    return query_nvidia_gpu_count()


def get_gpu_memory(dev: int) -> t.Tuple[float, float]:
    """
    Return Total Memory and Free Memory in given GPU device. in MiB
    """
    import pynvml.nvml  # type: ignore
    from pynvml.smi import nvidia_smi  # type: ignore

    unit_multiplier = {
        "PiB": 1024.0 * 1024 * 1024,
        "TiB": 1024.0 * 1024,
        "GiB": 1024.0,
        "MiB": 1.0,
        "KiB": 1.0 / 1024,
        "B": 1.0 / 1024 / 1024,
    }

    try:
        inst = nvidia_smi.getInstance()
        query: t.Dict[str, int] = inst.DeviceQuery(dev)  # type: ignore
    except (pynvml.nvml.NVMLError, OSError):
        return 0.0, 0.0

    try:
        gpus: t.List[t.Dict[str, t.Any]] = query.get("gpu", [])  # type: ignore
        gpu = gpus[dev]
        unit = gpu["fb_memory_usage"]["unit"]
        total = gpu["fb_memory_usage"]["total"] * unit_multiplier[unit]
        free = gpu["fb_memory_usage"]["free"] * unit_multiplier[unit]
        return total, free
    except IndexError:
        raise ValueError(f"Invalid GPU device index {dev}")
    except KeyError:
        raise RuntimeError(f"unexpected nvml query result: {query}")



from __future__ import annotations

import typing as t
import logging
from typing import TYPE_CHECKING

import bentoml
from bentoml import Tag
from bentoml.models import Model
from bentoml.models import ModelContext
from bentoml.exceptions import NotFound
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..models import PKL_EXT
from ..models import SAVE_NAMESPACE
from ..models.model import ModelSignature

if TYPE_CHECKING:
    from .. import external_typing as ext

    ModelType = t.Any

MODULE_NAME = "bentoml.picklable_model"

try:
    import cloudpickle  # type: ignore
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        f"""cloudpickle is required in order to use the module `{MODULE_NAME}`, install
         cloudpickle with `pip install cloudpickle`.
        """
    )


logger = logging.getLogger(__name__)


def get(tag_like: str | Tag) -> Model:
    model = bentoml.models.get(tag_like)
    if model.info.module not in (MODULE_NAME, __name__):
        raise NotFound(
            f"Model {model.tag} was saved with module {model.info.module}, failed loading with {MODULE_NAME}."
        )
    return model


def load_model(bento_model: str | Tag | Model) -> ModelType:
    """
    Load the picklable model with the given tag from the local BentoML model store.

    Args:
        bento_model (``str`` ``|`` :obj:`~bentoml.Tag` ``|`` :obj:`~bentoml.Model`):
            Either the tag of the model to get from the store, or a BentoML `~bentoml.Model`
            instance to load the model from.
        ...
    Returns:
        ``object``
            The picklable model loaded from the model store or BentoML :obj:`~bentoml.Model`.
    Example:
    .. code-block:: python
        import bentoml

        picklable_model = bentoml.picklable_model.load_model('my_model:latest')
    """  # noqa
    if not isinstance(bento_model, Model):
        bento_model = get(bento_model)

    if bento_model.info.module not in (MODULE_NAME, __name__):
        raise BentoMLException(
            f"Model {bento_model.tag} was saved with module {bento_model.info.module}, failed loading with {MODULE_NAME}."
        )
    model_file = bento_model.path_of(f"{SAVE_NAMESPACE}{PKL_EXT}")

    with open(model_file, "rb") as f:
        # The protocol version used is detected automatically, so we do not
        # have to specify it.
        return cloudpickle.load(f)


def save_model(
    name: str,
    model: ModelType,
    *,
    signatures: dict[str, ModelSignature] | None = None,
    labels: t.Dict[str, str] | None = None,
    custom_objects: t.Dict[str, t.Any] | None = None,
    metadata: t.Dict[str, t.Any] | None = None,
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (:code:`str`):
            Name for given model instance. This should pass Python identifier check.
        model (:code:`Union[BaseEstimator, Pipeline]`):
            Instance of model to be saved.
        signatures (:code: `Dict[str, ModelSignatureDict]`)
            Methods to expose for running inference on the target model. Signatures are
             used for creating Runner instances when serving model with bentoml.Service
        labels (:code:`Dict[str, str]`, `optional`, default to :code:`None`):
            user-defined labels for managing models, e.g. team=nlp, stage=dev
        custom_objects (:code:`Dict[str, Any]]`, `optional`, default to :code:`None`):
            user-defined additional python objects to be saved alongside the model,
             e.g. a tokenizer instance, preprocessor function, model configuration json
        metadata (:code:`Dict[str, Any]`, `optional`,  default to :code:`None`):
            Custom metadata for given model.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format `name:version` where `name` is
        the user-defined model's name, and a generated `version`.

    Examples:

    .. code-block:: python

        import bentoml

        tag = bentoml.picklable.save_model('picklable_pyobj', model)

        # load the model back:
        loaded = bentoml.picklable.load_model("picklable_pyobj:latest")
        # or:
        loaded = bentoml.picklable.load_model(tag)
    """  # noqa
    context = ModelContext(
        framework_name="cloudpickle",
        framework_versions={"cloudpickle": cloudpickle.__version__},
    )

    if signatures is None:
        logger.info(
            'Using default model signature `{"__call__": {"batchable": False}}` for picklable model'
        )
        signatures = {"__call__": ModelSignature(batchable=False)}

    with bentoml.models.create(
        name,
        module=MODULE_NAME,
        labels=labels,
        custom_objects=custom_objects,
        metadata=metadata,
        context=context,
        signatures=signatures,
    ) as _model:

        with open(_model.path_of(f"{SAVE_NAMESPACE}{PKL_EXT}"), "wb") as f:
            cloudpickle.dump(model, f)

        return _model.tag


def get_runnable(bento_model: Model):
    """
    Private API: use :obj:`~bentoml.Model.to_runnable` instead.
    """

    class PicklableRunnable(bentoml.Runnable):
        SUPPORT_NVIDIA_GPU = False  # type: ignore
        SUPPORT_CPU_MULTI_THREADING = False  # type: ignore

        def __init__(self):
            super().__init__()
            self.model = load_model(bento_model)

    def _get_run(method_name: str):
        def _run(
            self: PicklableRunnable,
            *args: ext.NpNDArray | ext.PdDataFrame,
            **kwargs: ext.NpNDArray | ext.PdDataFrame,
        ) -> ext.NpNDArray:
            assert isinstance(method_name, str), repr(method_name)
            return getattr(self.model, method_name)(*args, **kwargs)

        return _run

    for method_name, options in bento_model.info.signatures.items():
        assert isinstance(method_name, str), repr(method_name)
        PicklableRunnable.add_method(
            _get_run(method_name),
            name=method_name,
            batchable=options.batchable,
            batch_dim=options.batch_dim,
            input_spec=options.input_spec,
            output_spec=options.output_spec,
        )

    return PicklableRunnable



import os
import typing as t
from typing import TYPE_CHECKING

import joblib  # type: ignore[reportMissingTypeStubs]
from simple_di import inject
from simple_di import Provide

import bentoml
from bentoml import Tag
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..models import PKL_EXT
from ..models import TXT_EXT
from ..models import SAVE_NAMESPACE
from ..utils.pkg import get_pkg_version
from .common.model_runner import BaseModelRunner
from ..configuration.containers import BentoMLContainer

try:
    import lightgbm as lgb  # type: ignore[reportMissingTypeStubs]
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """lightgbm is required in order to use module `bentoml.lightgbm`, install
        lightgbm with `pip install lightgbm`. For more information, refer to
        https://github.com/microsoft/LightGBM/tree/master/python-package
        """
    )


if TYPE_CHECKING:
    from .. import external_typing as ext
    from ..models import ModelStore

    _LightGBMModelType = t.Union[
        "lgb.LGBMModel",
        "lgb.LGBMClassifier",
        "lgb.LGBMRegressor",
        "lgb.LGBMRanker",
    ]

MODULE_NAME = "bentoml.lightgbm"


def _get_model_info(
    tag: t.Union[str, Tag],
    booster_params: t.Optional[t.Dict[str, t.Union[str, int]]],
    model_store: "ModelStore",
) -> t.Tuple["Model", str, t.Dict[str, t.Any]]:
    model = model_store.get(tag)
    if model.info.module not in (MODULE_NAME, __name__):
        raise BentoMLException(
            f"Model {tag} was saved with module {model.info.module}, failed loading with {MODULE_NAME}."
        )
    _fname = (
        f"{SAVE_NAMESPACE}{TXT_EXT}"
        if not model.info.options["sklearn_api"]
        else f"{SAVE_NAMESPACE}{PKL_EXT}"
    )
    model_file = model.path_of(_fname)
    _booster_params: t.Dict[str, t.Union[str, int]] = (
        dict() if not booster_params else booster_params
    )
    for key, value in model.info.options.items():
        if key not in _booster_params:
            _booster_params[key] = value

    return model, model_file, _booster_params


@inject
def load(
    tag: t.Union[str, Tag],
    booster_params: t.Optional[t.Dict[str, t.Union[str, int]]] = None,
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> t.Union["lgb.basic.Booster", "_LightGBMModelType"]:
    """
    Load a model from BentoML local modelstore with given name.

    Args:
        tag (:code:`Union[str, Tag]`):
            Tag of a saved model in BentoML local modelstore.
        booster_params (:code:`Dict[str, Union[str, int]]`, `optional`, defaults to `None`):
            Parameters for boosters. Refers to `Parameters Docs <https://lightgbm.readthedocs.io/en/latest/Parameters.html>`_
            for more information.
        model_store (:mod:`~bentoml._internal.models.store.ModelStore`, default to :mod:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj::code:`Union[lightgbm.LGBMModel, lightgbm.LGBMClassifier, lightgbm.LGBMRegressor, lightgbm.LGBMRanker, lightgbm.basic.Booster]`: An instance of either
        :obj::code:`Union[lightgbm.LGBMModel, lightgbm.LGBMClassifier, lightgbm.LGBMRegressor, lightgbm.LGBMRanker]` or :obj:`lightgbm.basic.Booster`
        from BentoML modelstore.

    Examples:

    .. code-block:: python

        import bentoml
        gbm = bentoml.lightgbm.load("my_lightgbm_model:latest")
    """  # noqa
    _, _model_file, _booster_params = _get_model_info(tag, booster_params, model_store)

    if os.path.splitext(_model_file)[1] == PKL_EXT:
        return joblib.load(_model_file)
    else:
        return lgb.Booster(params=_booster_params, model_file=_model_file)


def save(
    name: str,
    model: t.Union["lgb.basic.Booster", "_LightGBMModelType"],
    *,
    booster_params: t.Optional[t.Dict[str, t.Union[str, int]]] = None,
    labels: t.Optional[t.Dict[str, str]] = None,
    custom_objects: t.Optional[t.Dict[str, t.Any]] = None,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (:code:`str`):
            Name for given model instance. This should pass Python identifier check.
        model (:code:`Union[lightgbm.basic.Booster, lightgbm.LGBMModel, lightgbm.LGBMClassifier, lightgbm.LGBMRegressor, lightgbm.LGBMRanker]`):
            Instance of model to be saved.
        booster_params (:code:`Dict[str, Union[str, int]]`, `optional`, defaults to `None`):
            Parameters for boosters. Refers to `Parameters Doc <https://lightgbm.readthedocs.io/en/latest/Parameters.html>`_
            for more information.
        labels (:code:`Dict[str, str]`, `optional`, default to :code:`None`):
            user-defined labels for managing models, e.g. team=nlp, stage=dev
        custom_objects (:code:`Dict[str, Any]]`, `optional`, default to :code:`None`):
            user-defined additional python objects to be saved alongside the model,
            e.g. a tokenizer instance, preprocessor function, model configuration json
        metadata (:code:`Dict[str, Any]`, `optional`, default to :code:`None`):
            Custom metadata for given model.
        model_store (:mod:`~bentoml._internal.models.store.ModelStore`, default to :mod:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format `name:version` where `name` is the user-defined model's name, and a generated `version` by BentoML.

    Examples:

    .. code-block:: python

        import bentoml

        import lightgbm as lgb
        import pandas as pd

        # load a dataset
        df_train = pd.read_csv("regression.train", header=None, sep="\t")
        df_test = pd.read_csv("regression.test", header=None, sep="\t")

        y_train = df_train[0]
        y_test = df_test[0]
        X_train = df_train.drop(0, axis=1)
        X_test = df_test.drop(0, axis=1)

        # create dataset for lightgbm
        lgb_train = lgb.Dataset(X_train, y_train)
        lgb_eval = lgb.Dataset(X_test, y_test, reference=lgb_train)

        # specify your configurations as a dict
        params = {
            "boosting_type": "gbdt",
            "objective": "regression",
            "metric": {"l2", "l1"},
            "num_leaves": 31,
            "learning_rate": 0.05,
        }

        # train
        gbm = lgb.train(
            params, lgb_train, num_boost_round=20, valid_sets=lgb_eval
        )

        tag = bentoml.lightgbm.save("my_lightgbm_model", gbm, booster_params=params)
    """  # noqa
    context: t.Dict[str, t.Any] = {
        "framework_name": "lightgbm",
        "pip_dependencies": [f"lightgbm=={get_pkg_version('lightgbm')}"],
    }

    with bentoml.models.create(
        name,
        module=MODULE_NAME,
        labels=labels,
        custom_objects=custom_objects,
        options=booster_params,
        context=context,
        metadata=metadata,
    ) as _model:

        _model.info.options["sklearn_api"] = False
        if any(
            isinstance(model, _)
            for _ in [
                lgb.LGBMModel,
                lgb.LGBMClassifier,
                lgb.LGBMRegressor,
                lgb.LGBMRanker,
            ]
        ):
            joblib.dump(model, _model.path_of(f"{SAVE_NAMESPACE}{PKL_EXT}"))
            _model.info.options["sklearn_api"] = True
        else:
            model.save_model(_model.path_of(f"{SAVE_NAMESPACE}{TXT_EXT}"))  # type: ignore

        return _model.tag


class _LightGBMRunner(BaseModelRunner):
    @inject
    def __init__(
        self,
        tag: t.Union[Tag, str],
        infer_api_callback: str,
        booster_params: t.Optional[t.Dict[str, t.Union[str, int]]],
        name: t.Optional[str] = None,
    ):
        super().__init__(tag, name=name)
        self._booster_params = booster_params if booster_params is not None else {}
        self._infer_api_callback = infer_api_callback

        self._predict_fn: t.Any = None
        self._model: t.Any = None

    def _is_gpu(self):
        try:
            if "device" in self._booster_params:
                assert isinstance(self._booster_params["device"], str)
                return "gpu" in self._booster_params["device"]
        except KeyError:
            return False

    @property
    def num_replica(self) -> int:
        if self._is_gpu() and self.resource_quota.nvidia_gpu:
            return self.resource_quota.nvidia_gpu
        return 1

    def _setup(self) -> None:
        self._model = load(
            tag=self._tag,
            booster_params=self._booster_params,
            model_store=self.model_store,
        )
        self._predict_fn = getattr(self._model, self._infer_api_callback)

    def _run_batch(self, input_data: "ext.NpNDArray") -> "ext.NpNDArray":  # type: ignore[reportIncompatibleMethodOverride]
        return self._predict_fn(input_data)


@inject
def load_runner(
    tag: t.Union[str, Tag],
    infer_api_callback: str = "predict",
    *,
    booster_params: t.Optional[t.Dict[str, t.Union[str, int]]] = None,
    name: t.Optional[str] = None,
) -> "_LightGBMRunner":
    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. :func:`bentoml.lightgbm.load_runner` implements a Runner class that
    wrap around a LightGBM model, which optimize it for the BentoML runtime.

    Args:
        tag (:code:`Union[str, Tag]`):
            Tag of a saved model in BentoML modelstore.
        infer_api_callback (:code:`str`, `optional`, default to :code:`predict`):
            Inference API callback from given model. If not specified, BentoML will use default :code:`predict`.
            Users can also choose to use :code:`predict_proba` for supported model.
        booster_params (:code:`Dict[str, Union[str, int]]`, `optional`, defaults to `None`):
            Parameters for boosters. Refers to `Parameters Doc <https://lightgbm.readthedocs.io/en/latest/Parameters.html>`_
            for more information.

    Returns:
        :obj:`~bentoml._internal.runner.Runner`: Runner instances for :mod:`bentoml.lightgbm` model

    Examples:

    .. code-block:: python

        import bentoml

        runner = bentoml.lightgbm.load_runner("my_lightgbm_model:latest")
        runner.run_batch(X_test, num_iteration=gbm.best_iteration)
    """
    return _LightGBMRunner(
        tag=tag,
        infer_api_callback=infer_api_callback,
        name=name,
        booster_params=booster_params,
    )



from __future__ import annotations

import os
import sys
import uuid
import random
import socket
import typing as t
import functools
import contextlib
from typing import overload
from typing import TYPE_CHECKING
from pathlib import Path
from datetime import date
from datetime import time
from datetime import datetime
from datetime import timedelta

import fs
import attr
import fs.copy

if sys.version_info >= (3, 8):
    from functools import cached_property
else:
    from backports.cached_property import cached_property

from .cattr import bentoml_cattr
from ..types import LazyType
from ..types import PathType
from ..types import MetadataDict
from ..types import MetadataType
from .lazy_loader import LazyLoader

if TYPE_CHECKING:
    from fs.base import FS

    P = t.ParamSpec("P")
    GenericFunction = t.Callable[P, t.Any]


C = t.TypeVar("C")
T = t.TypeVar("T")
_T_co = t.TypeVar("_T_co", covariant=True, bound=t.Any)


__all__ = [
    "bentoml_cattr",
    "cached_property",
    "cached_contextmanager",
    "reserve_free_port",
    "catch_exceptions",
    "LazyLoader",
    "validate_or_create_dir",
    "display_path_under_home",
]


@overload
def kwargs_transformers(
    func: GenericFunction[t.Concatenate[str, bool, t.Iterable[str], P]],
    *,
    transformer: GenericFunction[t.Any],
) -> GenericFunction[t.Concatenate[str, t.Iterable[str], bool, P]]:
    ...


@overload
def kwargs_transformers(
    func: None = None, *, transformer: GenericFunction[t.Any]
) -> GenericFunction[t.Any]:
    ...


def kwargs_transformers(
    _func: t.Callable[..., t.Any] | None = None,
    *,
    transformer: GenericFunction[t.Any],
) -> GenericFunction[t.Any]:
    def decorator(func: GenericFunction[t.Any]) -> t.Callable[P, t.Any]:
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> t.Any:
            return func(*args, **{k: transformer(v) for k, v in kwargs.items()})

        return wrapper

    if _func is None:
        return decorator
    return decorator(_func)


@t.overload
def first_not_none(*args: T | None, default: T) -> T:
    ...


@t.overload
def first_not_none(*args: T | None) -> T | None:
    ...


def first_not_none(*args: T | None, default: None | T = None) -> T | None:
    """
    Returns the first argument that is not None.
    """
    return next((arg for arg in args if arg is not None), default)


def randomize_runner_name(module_name: str):
    return f"{module_name.split('.')[-1]}_{uuid.uuid4().hex[:6].lower()}"


def validate_or_create_dir(*path: PathType) -> None:
    for p in path:
        path_obj = Path(p)

        if path_obj.exists():
            if not path_obj.is_dir():
                raise OSError(20, f"{path_obj} is not a directory")
        else:
            path_obj.mkdir(parents=True)


def calc_dir_size(path: PathType) -> int:
    return sum(f.stat().st_size for f in Path(path).glob("**/*") if f.is_file())


def display_path_under_home(path: str) -> str:
    # Shorten path under home directory with leading `~`
    # e.g. from `/Users/foo/bar` to just `~/bar`
    try:
        return str("~" / Path(path).relative_to(Path.home()))
    except ValueError:
        # when path is not under home directory, return original full path
        return path


def human_readable_size(size: t.Union[int, float], decimal_places: int = 2) -> str:
    for unit in ["B", "KiB", "MiB", "GiB", "TiB", "PiB"]:
        if size < 1024.0 or unit == "PiB":
            break
        size /= 1024.0
    else:
        raise ValueError("size is too large")
    return f"{size:.{decimal_places}f} {unit}"


class catch_exceptions(t.Generic[_T_co], object):
    def __init__(
        self,
        catch_exc: t.Union[t.Type[BaseException], t.Tuple[t.Type[BaseException], ...]],
        throw_exc: t.Callable[[str], BaseException],
        msg: str = "",
        fallback: t.Optional[_T_co] = None,
        raises: t.Optional[bool] = True,
    ) -> None:
        self._catch_exc = catch_exc
        self._throw_exc = throw_exc
        self._msg = msg
        self._fallback = fallback
        self._raises = raises

    def __call__(self, func: t.Callable[P, _T_co]) -> t.Callable[P, t.Optional[_T_co]]:
        @functools.wraps(func)
        def _(*args: P.args, **kwargs: P.kwargs) -> t.Optional[_T_co]:
            try:
                return func(*args, **kwargs)
            except self._catch_exc:
                if self._raises:
                    raise self._throw_exc(self._msg)
                return self._fallback

        return _


@contextlib.contextmanager
def reserve_free_port(
    host: str = "localhost",
    prefix: t.Optional[str] = None,
    max_retry: int = 50,
) -> t.Iterator[int]:
    """
    detect free port and reserve until exit the context
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    if prefix is not None:
        prefix_num = int(prefix) * 10 ** (5 - len(prefix))
        suffix_range = min(65535 - prefix_num, 10 ** (5 - len(prefix)))
        for _ in range(max_retry):
            suffix = random.randint(0, suffix_range)
            port = int(f"{prefix_num + suffix}")
            try:
                sock.bind((host, port))
                break
            except OSError:
                continue
        else:
            raise RuntimeError(
                f"cannot find free port with prefix {prefix} after {max_retry} retries"
            )
    else:
        sock.bind((host, 0))
    port = sock.getsockname()[1]
    yield port
    sock.close()


def copy_file_to_fs_folder(
    src_path: str,
    dst_fs: FS,
    dst_folder_path: str = ".",
    dst_filename: t.Optional[str] = None,
):
    """Copy the given file at src_path to dst_fs filesystem, under its dst_folder_path
    folder with dst_filename as file name. When dst_filename is None, keep the original
    file name.
    """
    src_path = os.path.realpath(os.path.expanduser(src_path))
    dir_name, file_name = os.path.split(src_path)
    src_fs = fs.open_fs(dir_name)
    dst_filename = file_name if dst_filename is None else dst_filename
    dst_path = fs.path.join(dst_folder_path, dst_filename)
    fs.copy.copy_file(src_fs, file_name, dst_fs, dst_path)


def resolve_user_filepath(filepath: str, ctx: t.Optional[str]) -> str:
    """Resolve the abspath of a filepath provided by user, which may contain "~" or may
    be a relative path base on ctx dir.
    """
    # Return if filepath exist after expanduser
    _path = os.path.expanduser(filepath)
    if os.path.exists(_path):
        return os.path.realpath(_path)

    # Try finding file in ctx if provided
    if ctx:
        _path = os.path.expanduser(os.path.join(ctx, filepath))
        if os.path.exists(_path):
            return os.path.realpath(_path)

    raise FileNotFoundError(f"file {filepath} not found")


def label_validator(
    _: t.Any, _attr: attr.Attribute[dict[str, str]], labels: dict[str, str]
):
    validate_labels(labels)


def validate_labels(labels: dict[str, str]):
    if not isinstance(labels, dict):
        raise ValueError("labels must be a dict!")

    for key, val in labels.items():
        if not isinstance(key, str):
            raise ValueError("label keys must be strings")

        if not isinstance(val, str):
            raise ValueError("label values must be strings")


def metadata_validator(
    _: t.Any, _attr: attr.Attribute[MetadataDict], metadata: MetadataDict
):
    validate_metadata(metadata)


def validate_metadata(metadata: MetadataDict):
    if not isinstance(metadata, dict):
        raise ValueError("metadata must be a dict!")

    for key, val in metadata.items():
        if not isinstance(key, (str, int, float)):
            raise ValueError("metadata keys must be strings")

        metadata[key] = _validate_metadata_entry(val)


def _validate_metadata_entry(entry: MetadataType) -> MetadataType:
    if isinstance(entry, dict):
        validate_metadata(entry)
    elif isinstance(entry, list):
        for i, val in enumerate(entry):
            entry[i] = _validate_metadata_entry(val)
    elif isinstance(entry, tuple):
        entry = tuple((_validate_metadata_entry(x) for x in entry))

    elif LazyType("numpy", "ndarray").isinstance(entry):
        entry = entry.tolist()  # type: ignore (LazyType)
        _validate_metadata_entry(entry)
    elif LazyType("numpy", "generic").isinstance(entry):
        entry = entry.item()  # type: ignore (LazyType)
        _validate_metadata_entry(entry)
    elif LazyType("scipy.sparse", "spmatrix").isinstance(entry):
        raise ValueError(
            "SciPy sparse matrices are currently not supported as metadata items; consider using a dictionary instead"
        )
    elif LazyType("pandas", "Series").isinstance(entry):
        entry = {entry.name: entry.to_dict()}
        _validate_metadata_entry(entry)
    elif LazyType("pandas.api.extensions", "ExtensionArray").isinstance(entry):
        entry = entry.to_numpy().tolist()  # type: ignore (LazyType)
        _validate_metadata_entry(entry)
    elif LazyType("pandas", "DataFrame").isinstance(entry):
        entry = entry.to_dict()  # type: ignore (LazyType)
        validate_metadata(entry)  # type: ignore (LazyType)
    elif LazyType("pandas", "Timestamp").isinstance(entry):
        entry = entry.to_pydatetime()  # type: ignore (LazyType)
    elif LazyType("pandas", "Timedelta").isinstance(entry):
        entry = entry.to_pytimedelta()  # type: ignore (LazyType)
    elif LazyType("pandas", "Period").isinstance(entry):
        entry = entry.to_timestamp().to_pydatetime()  # type: ignore (LazyType)
    elif LazyType("pandas", "Interval").isinstance(entry):
        entry = (entry.left, entry.right)  # type: ignore (LazyType)
        _validate_metadata_entry(entry)
    elif not isinstance(
        entry, (str, bytes, bool, int, float, complex, datetime, date, time, timedelta)
    ):
        raise ValueError(
            f"metadata entries must be basic python types like 'str', 'int', or 'complex', got '{type(entry).__name__}'"
        )

    return entry


VT = t.TypeVar("VT")


class cached_contextmanager:
    """
    Just like contextlib.contextmanager, but will cache the yield value for the same
    arguments. When one instance of the contextmanager exits, the cache value will
    also be poped.
    Example Usage:
    (To reuse the container based on the same image)
    >>> @cached_contextmanager("{docker_image.id}")
    >>> def start_docker_container_from_image(docker_image, timeout=60):
    >>>     container = ...
    >>>     yield container
    >>>     container.stop()
    """

    def __init__(self, cache_key_template: t.Optional[str] = None):
        self._cache_key_template = cache_key_template
        self._cache: t.Dict[t.Any, t.Any] = {}

    def __call__(
        self, func: "t.Callable[P, t.Generator[VT, None, None]]"
    ) -> "t.Callable[P, t.ContextManager[VT]]":

        func_m = contextlib.contextmanager(func)

        @contextlib.contextmanager
        @functools.wraps(func)
        def _func(*args: "P.args", **kwargs: "P.kwargs") -> t.Any:
            import inspect

            bound_args = inspect.signature(func).bind(*args, **kwargs)
            bound_args.apply_defaults()
            if self._cache_key_template:
                cache_key = self._cache_key_template.format(**bound_args.arguments)
            else:
                cache_key = repr(tuple(bound_args.arguments.values()))
            if cache_key in self._cache:
                yield self._cache[cache_key]
            else:
                with func_m(*args, **kwargs) as value:
                    self._cache[cache_key] = value
                    yield value
                    self._cache.pop(cache_key)

        return _func



from __future__ import annotations

import typing as t
import logging

import click

logger = logging.getLogger(__name__)


def unparse_click_params(
    params: dict[str, t.Any],
    command_params: list[click.Parameter],
    *,
    factory: t.Callable[..., t.Any] | None = None,
) -> list[str]:
    """
    Unparse click call to a list of arguments. Used to modify some parameters and
    restore to system command. The goal is to unpack cases where parameters can be parsed multiple times.

    Refers to ./buildx.py for examples of this usage. This is also used to unparse parameters for running API server.

    Args:
        params (`dict[str, t.Any]`):
            The dictionary of the parameters that is parsed from click.Context.
        command_params (`list[click.Parameter]`):
            The list of paramters (Arguments/Options) that is part of a given command.

    Returns:
        Unparsed list of arguments that can be redirected to system commands.

    Implementation:
        For cases where options is None, or the default value is None, we will remove it from the first params list.
        Currently it doesn't support unpacking `prompt_required` or `confirmation_prompt`.
    """
    args: list[str] = []

    # first filter out all parsed parameters that have value None
    # This means that the parameter was not set by the user
    params = {k: v for k, v in params.items() if v not in [None, (), []]}

    for command_param in command_params:
        if isinstance(command_param, click.Argument):
            # Arguments.nargs, Arguments.required
            if command_param.name in params:
                if command_param.nargs > 1:
                    # multiple arguments are passed as a list.
                    # In this case we try to convert all None to an empty string
                    args.extend(
                        list(
                            filter(
                                lambda x: "" if x is None else x,
                                params[command_param.name],
                            )
                        )
                    )
                else:
                    args.append(params[command_param.name])
        elif isinstance(command_param, click.Option):
            if command_param.name in params:
                if (
                    command_param.confirmation_prompt
                    or command_param.prompt is not None
                ):
                    logger.warning(
                        f"{command_params} is a prompt, skip parsing it for now."
                    )
                    pass
                if command_param.is_flag:
                    args.append(command_param.opts[-1])
                else:
                    cmd = f"--{command_param.name.replace('_','-')}"
                    if command_param.multiple:
                        for var in params[command_param.name]:
                            args.extend([cmd, var])
                    else:
                        args.extend([cmd, params[command_param.name]])
        else:
            logger.warning(
                "Given command params is a subclass of click.Parameter, but not a click.Argument or click.Option. Passing through..."
            )

    # We will also convert values if factory is parsed:
    if factory is not None:
        return list(map(factory, args))

    return args



import os
import re
import time
import typing as t
import difflib
import logging
import functools
from typing import TYPE_CHECKING

import click
from click import ClickException
from click.exceptions import UsageError

from ...exceptions import BentoMLException
from ..configuration import CONFIG_ENV_VAR
from ..configuration import set_debug_mode
from ..configuration import load_global_config
from ..utils.analytics import track
from ..utils.analytics import CliEvent
from ..utils.analytics import cli_events_map
from ..utils.analytics import BENTOML_DO_NOT_TRACK

if TYPE_CHECKING:
    P = t.ParamSpec("P")

    class ClickFunctionWrapper(t.Protocol[P]):
        __name__: str
        __click_params__: t.List[click.Option]

        def __call__(*args: P.args, **kwargs: P.kwargs) -> t.Callable[P, t.Any]:
            ...

    WrappedCLI = t.Callable[P, ClickFunctionWrapper[t.Any]]


logger = logging.getLogger(__name__)


class BentoMLCommandGroup(click.Group):
    """Click command class customized for BentoML CLI, allow specifying a default
    command for each group defined
    """

    NUMBER_OF_COMMON_PARAMS = 4

    @staticmethod
    def bentoml_common_params(
        func: "t.Callable[P, t.Any]",
    ) -> "WrappedCLI[bool, bool, t.Optional[str]]":
        # NOTE: update NUMBER_OF_COMMON_PARAMS when adding option.

        @click.option(
            "-q",
            "--quiet",
            is_flag=True,
            default=False,
            help="Suppress all warnings and info logs",
        )
        @click.option(
            "--verbose",
            "--debug",
            is_flag=True,
            default=False,
            help="Generate debug information",
        )
        @click.option(
            "--do-not-track",
            is_flag=True,
            default=False,
            envvar=BENTOML_DO_NOT_TRACK,
            help="Do not send usage info",
        )
        @click.option(
            "--config",
            type=click.Path(exists=True),
            envvar=CONFIG_ENV_VAR,
            help="BentoML configuration YAML file to apply",
        )
        @functools.wraps(func)
        def wrapper(
            quiet: bool,
            verbose: bool,
            config: t.Optional[str],
            *args: "P.args",
            **kwargs: "P.kwargs",
        ) -> t.Any:
            if config:
                load_global_config(config)

            if quiet:
                logging.getLogger("bentoml").setLevel(logging.ERROR)
                if verbose:
                    logger.warning(
                        "The bentoml command option `--verbose/--debug` is ignored when"
                        "the `--quiet` flag is also in use"
                    )
            elif verbose:
                set_debug_mode(True)

            return func(*args, **kwargs)

        return wrapper

    @staticmethod
    def bentoml_track_usage(
        func: t.Union["t.Callable[P, t.Any]", "ClickFunctionWrapper[t.Any]"],
        cmd_group: click.Group,
        **kwargs: t.Any,
    ):
        command_name = kwargs.get("name", func.__name__)

        @functools.wraps(func)
        def wrapper(do_not_track: bool, *args: "P.args", **kwargs: "P.kwargs") -> t.Any:
            if do_not_track:
                os.environ[BENTOML_DO_NOT_TRACK] = str(True)
                return func(*args, **kwargs)

            start_time = time.time_ns()

            if (
                cmd_group.name in cli_events_map
                and command_name in cli_events_map[cmd_group.name]
            ):
                get_tracking_event = functools.partial(
                    cli_events_map[cmd_group.name][command_name],
                    cmd_group.name,
                    command_name,
                )
            else:

                def get_tracking_event(ret: t.Any) -> CliEvent:
                    return CliEvent(
                        cmd_group=cmd_group.name,
                        cmd_name=command_name,
                    )

            try:
                return_value = func(*args, **kwargs)
                event = get_tracking_event(return_value)
                duration_in_ns = time.time_ns() - start_time
                event.duration_in_ms = duration_in_ns / 1e6
                track(event)
                return return_value
            except BaseException as e:
                event = get_tracking_event(None)
                duration_in_ns = time.time_ns() - start_time
                event.duration_in_ms = duration_in_ns / 1e6
                event.error_type = type(e).__name__
                event.return_code = 2 if isinstance(e, KeyboardInterrupt) else 1
                track(event)
                raise

        return wrapper

    @staticmethod
    def raise_click_exception(
        func: t.Union["t.Callable[P, t.Any]", "ClickFunctionWrapper[t.Any]"],
        cmd_group: click.Group,
        **kwargs: t.Any,
    ) -> "ClickFunctionWrapper[t.Any]":
        command_name = kwargs.get("name", func.__name__)

        @functools.wraps(func)
        def wrapper(*args: "P.args", **kwargs: "P.kwargs") -> t.Any:
            try:
                return func(*args, **kwargs)
            except BentoMLException as err:
                msg = f"[{cmd_group.name}] `{command_name}` failed: {str(err)}"
                raise ClickException(click.style(msg, fg="red")) from err

        return t.cast("ClickFunctionWrapper[t.Any]", wrapper)

    def command(
        self, *args: t.Any, **kwargs: t.Any
    ) -> "t.Callable[[t.Callable[P, t.Any]], click.Command]":
        if "context_settings" not in kwargs:
            kwargs["context_settings"] = {}
        kwargs["context_settings"]["max_content_width"] = 120

        def wrapper(func: "t.Callable[P, t.Any]") -> click.Command:
            # add common parameters to command.
            func = BentoMLCommandGroup.bentoml_common_params(func)
            # Send tracking events before command finish.
            func = BentoMLCommandGroup.bentoml_track_usage(func, self, **kwargs)
            # If BentoMLException raise ClickException instead before exit.
            func = BentoMLCommandGroup.raise_click_exception(func, self, **kwargs)

            # move common parameters to end of the parameters list
            func.__click_params__ = (
                func.__click_params__[-self.NUMBER_OF_COMMON_PARAMS :]
                + func.__click_params__[: -self.NUMBER_OF_COMMON_PARAMS]
            )
            return super(BentoMLCommandGroup, self).command(*args, **kwargs)(func)

        return wrapper

    def resolve_command(
        self, ctx: click.Context, args: t.List[str]
    ) -> t.Tuple[str, click.Command, t.List[str]]:
        try:
            return super(BentoMLCommandGroup, self).resolve_command(ctx, args)
        except UsageError as e:
            error_msg = str(e)
            original_cmd_name = click.utils.make_str(args[0])
            matches = difflib.get_close_matches(
                original_cmd_name, self.list_commands(ctx), 3, 0.5
            )
            if matches:
                fmt_matches = "\n    ".join(matches)
                error_msg += "\n\n"
                error_msg += f"Did you mean?\n    {fmt_matches}"
            raise UsageError(error_msg, e.ctx)


def is_valid_bento_tag(value: str) -> bool:
    return re.match(r"^[A-Za-z_][A-Za-z_0-9]*:[A-Za-z0-9.+-_]*$", value) is not None


def is_valid_bento_name(value: str) -> bool:
    return re.match(r"^[A-Za-z_0-9]*$", value) is not None



# type: ignore[reportUnusedFunction]
import logging

import click

from bentoml.exceptions import CLIException

from ..cli.click_utils import BentoMLCommandGroup
from ..yatai_rest_api_client.yatai import YataiRESTApiClient
from ..yatai_rest_api_client.config import add_context
from ..yatai_rest_api_client.config import YataiClientContext
from ..yatai_rest_api_client.config import default_context_name

logger = logging.getLogger(__name__)


def add_login_command(cli: click.Group) -> None:
    @cli.group(name="yatai", cls=BentoMLCommandGroup)
    def yatai_cli():
        """Yatai Subcommands Groups"""

    @yatai_cli.command(help="Login to Yatai server")
    @click.option(
        "--endpoint", type=click.STRING, help="Yatai endpoint, i.e: https://yatai.com"
    )
    @click.option("--api-token", type=click.STRING, help="Yatai user API token")
    def login(endpoint: str, api_token: str) -> None:
        if not endpoint:
            raise CLIException("need --endpoint")

        if not api_token:
            raise CLIException("need --api-token")

        yatai_rest_client = YataiRESTApiClient(endpoint, api_token)
        user = yatai_rest_client.get_current_user()

        if user is None:
            raise CLIException("current user is not found")

        org = yatai_rest_client.get_current_organization()

        if org is None:
            raise CLIException("current organization is not found")

        ctx = YataiClientContext(
            name=default_context_name,
            endpoint=endpoint,
            api_token=api_token,
            email=user.email,
        )

        add_context(ctx)

        logger.info(
            f"login successfully! user: {user.name}, organization: {org.name}",
        )



import os
import shutil
import typing as t
from typing import TYPE_CHECKING

from simple_di import inject
from simple_di import Provide

import bentoml
from bentoml import Tag
from bentoml.exceptions import BentoMLException
from bentoml.exceptions import MissingDependencyException

from ..models import SAVE_NAMESPACE
from .common.model_runner import BaseModelRunner
from ..configuration.containers import BentoMLContainer

if TYPE_CHECKING:
    from .. import external_typing as ext
    from ..models import ModelStore

try:
    from PyRuntime import __spec__ as _spec
    from PyRuntime import ExecutionSession
except ImportError:  # pragma: no cover
    raise MissingDependencyException(
        """\
PyRuntime is not found in PYTHONPATH. Refers to
 https://github.com/onnx/onnx-mlir#installation-on-unix for
 more information.
    """
    )

ONNXMLIR_EXTENSION: str = ".so"

MODULE_NAME = "bentoml.onnxmlir"


@inject
def load(
    tag: t.Union[str, Tag],
    model_store: "ModelStore" = Provide[BentoMLContainer.model_store],
) -> "ExecutionSession":
    """
    Load a model from BentoML local modelstore with given name.

    onnx-mlir is a compiler technology that can take an onnx model and lower it
    (using llvm) to an inference library that is optimized and has little external
    dependencies.

    The PyRuntime interface is created during the build of onnx-mlir using pybind.
    See the onnx-mlir supporting documentation for detail.

    Args:
        tag (:code:`Union[str, Tag]`):
            Tag of a saved model in BentoML local modelstore.
        model_store (:mod:`~bentoml._internal.models.store.ModelStore`, default to :mod:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj:`ExecutionSession`: an instance of ONNX-MLir compiled model from BentoML modelstore.

    Examples:

    .. code-block:: python

        import bentoml

        session = bentoml.onnxmlir.load(tag)
        session.run(data)

    """
    model = model_store.get(tag)
    if model.info.module not in (MODULE_NAME, __name__):
        raise BentoMLException(
            f"Model {tag} was saved with module {model.info.module}, failed loading with {MODULE_NAME}."
        )
    compiled_path = model.path_of(model.info.options["compiled_path"])
    return ExecutionSession(compiled_path, "run_main_graph")  # type: ignore


def save(
    name: str,
    model: t.Any,
    *,
    labels: t.Optional[t.Dict[str, str]] = None,
    custom_objects: t.Optional[t.Dict[str, t.Any]] = None,
    metadata: t.Optional[t.Dict[str, t.Any]] = None,
) -> Tag:
    """
    Save a model instance to BentoML modelstore.

    Args:
        name (:code:`str`):
            Name for given model instance. This should pass Python identifier check.
        model (:code:`str`):
            Path to compiled model by MLIR.
        labels (:code:`Dict[str, str]`, `optional`, default to :code:`None`):
            user-defined labels for managing models, e.g. team=nlp, stage=dev
        custom_objects (:code:`Dict[str, Any]]`, `optional`, default to :code:`None`):
            user-defined additional python objects to be saved alongside the model,
            e.g. a tokenizer instance, preprocessor function, model configuration json
        metadata (:code:`Dict[str, Any]`, `optional`,  default to :code:`None`):
            Custom metadata for given model.
        model_store (:mod:`~bentoml._internal.models.store.ModelStore`, default to :mod:`BentoMLContainer.model_store`):
            BentoML modelstore, provided by DI Container.

    Returns:
        :obj:`~bentoml.Tag`: A :obj:`tag` with a format `name:version` where `name` is the user-defined model's name, and a generated `version` by BentoML.

    Examples:

    .. code-block:: python

        import sys
        import os
        import subprocess

        import bentoml
        import tensorflow as tf

        sys.path.append("/workdir/onnx-mlir/build/Debug/lib/")

        from PyRuntime import ExecutionSession

        class NativeModel(tf.Module):
            def __init__(self):
                super().__init__()
                self.weights = np.asfarray([[1.0], [1.0], [1.0], [1.0], [1.0]])
                self.dense = lambda inputs: tf.matmul(inputs, self.weights)

            @tf.function(
                input_signature=[tf.TensorSpec(shape=[1, 5], dtype=tf.float64, name="inputs")]
            )
            def __call__(self, inputs):
                return self.dense(inputs)

        directory = "/tmp/model"
        model = NativeModel()
        tf.saved_model.save(model, directory)

        model_path = os.path.join(directory, "model.onnx")
        command = [
            "python",
            "-m",
            "tf2onnx.convert",
            "--saved-model",
            directory,
            "--output",
            model_path,
        ]
        docker_proc = subprocess.Popen(  # noqa
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=tmpdir, text=True
        )
        stdout, stderr = docker_proc.communicate()

        sys.path.append("/workdir/onnx-mlir/build/Debug/lib/")
        model_location = os.path.join(directory, "model.onnx")
        command = ["./onnx-mlir", "--EmitLib", model_location]
        onnx_mlir_loc = "/workdir/onnx-mlir/build/Debug/bin"

        docker_proc = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
            cwd=onnx_mlir_loc,
        )
        stdout, stderr = docker_proc.communicate()

        model_path = os.path.join(directory, "model.so")
        tag = bentoml.onnxmlir.save("compiled_model", model)

    """
    context: t.Dict[str, t.Any] = {
        "framework_name": "onnxmlir",
        "onnxmlir_version": _spec.origin,
    }

    with bentoml.models.create(
        name,
        module=MODULE_NAME,
        labels=labels,
        custom_objects=custom_objects,
        options=None,
        metadata=metadata,
        context=context,
    ) as _model:
        fpath = _model.path_of(f"{SAVE_NAMESPACE}{ONNXMLIR_EXTENSION}")
        _model.info.options["compiled_path"] = os.path.relpath(fpath, _model.path)
        shutil.copyfile(model, fpath)

        return _model.tag


class _ONNXMLirRunner(BaseModelRunner):
    @property
    def num_replica(self) -> int:
        return 1

    def _setup(self) -> None:
        self._session = load(self._tag, model_store=self.model_store)

    def _run_batch(self, input_data: "ext.NpNDArray") -> "ext.NpNDArray":  # type: ignore
        return self._session.run(input_data)  # type: ignore


def load_runner(
    tag: t.Union[str, Tag],
    *,
    name: t.Optional[str] = None,
) -> "_ONNXMLirRunner":
    """
    Runner represents a unit of serving logic that can be scaled horizontally to
    maximize throughput. :func:`bentoml.onnxmlir.load_runner` implements a Runner class that
    wrap around a ONNX-MLir compiled model, which optimize it for the BentoML runtime.

    Args:
        tag (:code:`Union[str, Tag]`):
            Tag of a saved model in BentoML local modelstore.

    Returns:
        :obj:`~bentoml._internal.runner.Runner`: Runner instances for :mod:`bentoml.xgboost` model

    Examples:

    .. code-block:: python

        import bentoml

        runner = bentoml.onnxmlir.load_runner(tag)
        res = runner.run_batch(pd_dataframe.to_numpy().astype(np.float64))

    """
    return _ONNXMLirRunner(tag=tag, name=name)
